
;*****************************************************************
;
; SHA-1 sobre SIMD - Versión Ensamblador Escalar.
;
; Pino, Jorge Daniel - JDanielPino@gmail.com
;
;*****************************************************************

; void get_sha1_asm(unsigned char* buffer, unsigned int size, unsigned int h[5]);
global get_sha1_asm
;
; Los parámetros de la función 'get_sha1_asm' se reciben en los
; siguientes registros:
;
; RDI:  'buffer'
; RSI:  'size'
; RDX:  'h[5]'
;

;----------------------------------------------------------------
;
; Definiciones de macros
;

;
; Asignamos aliases a los registros donde se reciben los
; parámetros, de manera tal que podamos referenciarlos con los
; nombres originales de esos parámetros. Obs: la macro 'xdefine',
; a diferencia de 'define', evita la redefinición.
%xdefine BUFFER_PTR    RDI
%xdefine END_OF_BUFFER RSI

;
; Vector h[5] donde se almacenará el hash calculado.
;
%xdefine H0 RDX
%xdefine H1 RDX + 4
%xdefine H2 RDX + 8
%xdefine H3 RDX + 12
%xdefine H4 RDX + 16

;
; Utilizamos a registros de propósito general para mantener las
; variables A, B, C, D y E del algoritmo de hashing.
;
%xdefine A R11D
%xdefine B R12D
%xdefine C R13D
%xdefine D R14D
%xdefine E R15D

;
; Funciones F1, F2, F3 y F4 utilizadas en los cálculos
; intermedios del hash. Devuelven el resultado en EAX.
; Parámetros:
;
; %1: b
; %2: c
; %3: d
;
;
; F1 = (b & c) | ((~b) & d)
%macro F1 3
  MOV EAX, %2   ; EAX = c
  MOV EBX, %1   ; EBX = b
  AND EAX, EBX  ; EAX = (b & c)
  NOT EBX       ; EBX = ~b
  AND EBX, %3   ; EBX = ((~b) & d)
  OR  EAX, EBX  ; EAX = (b & c) | ((~b) & d)
%endmacro

;
; F2 = b ^ c ^ d;
%macro F2 3
  MOV EAX, %1  ; EAX = b
  XOR EAX, %2  ; EAX = b ^ c
  XOR EAX, %3  ; EAX = b ^ c ^d
%endmacro

;
; F3 = (b & c) | (b & d) | (c & d)
%macro F3 3
  MOV EAX, %1   ; EAX = b
  AND EAX, %2   ; EAX = b & c
  MOV EBX, %1   ; EBX = b
  AND EBX, %3   ; EBX = b & d
  OR  EAX, EBX  ; EAX = (b & c) | (b & d)
  MOV EBX, %2   ; EBX = c
  AND EBX, %3   ; EBX = c & d
  OR  EAX, EBX  ; EAX = (b & c) | (b & d) | (c & d)
%endmacro

;
; F4 = F2 = b ^ c ^ d;
%define F4 F2

;
; Constantes K.
;
%xdefine K1 0x5A827999
%xdefine K2 0x6ED9EBA1
%xdefine K3 0x8F1BBCDC
%xdefine K4 0xCA62C1D6
%xdefine K  K1

;
; Parámetros:
;
; %1: número de ronda (de 0 a 79).
; %2: puntero al buffer.
; %3: A
; %4: B
; %5: C
; %6: D
; %7: E
;
%macro UPDATE_ABCDE 7

  %if (%1 < 16)
    MOV EAX, [%2 + 4*%1]

    ;
    ; Formateamos el valor a 'Little Endian' y lo guardamos en
    ; en la posición correspondiente del vector 'w'.
    BSWAP EAX
    MOV [w + 4*%1], EAX  ; w[i] = EAX

  %elif
    ;
    ; w[i] = ROTATE_LEFT((w[i-3] ^ w[i-8] ^ w[i-14] ^ w[i-16]), 1);
    ;
    MOV EAX, [w + (%1-3)*4]   ; EAX = w[i-3]
    XOR EAX, [w + (%1-8)*4]   ; EAX = w[i-3] XOR w[i-8]
    XOR EAX, [w + (%1-14)*4]  ; EAX = w[i-3] XOR w[i-8] XOR w[j-14]
    XOR EAX, [w + (%1-16)*4]  ; EAX = w[i-3] XOR w[i-8] XOR w[i-14] XOR w[i-16]
    ROL EAX, 1                ; EAX = ROTATE_LEFT(w[i-3] XOR w[i-8] XOR w[i-14] XOR w[i-16], 1)
    MOV [w + 4*%1], EAX       ; w[i] = EAX
  %endif

  %if (%1 <= 19)
    F1 %4, %5, %6
    %xdefine K K1

  %elif (%1 <= 39)
    F2 %4, %5, %6
    %xdefine K K2

  %elif (%1 <= 59)
    F3 %4, %5, %6
    %xdefine K K3

  %else
    F4 %4, %5, %6
    %xdefine K K4
  %endif

  ;
  ; Actualizamos la variable 'E', que en la próxima ronda será
  ; 'A'.
  ;
  ; E = E + f + ROTATE_LEFT(A, 5) + w[i] + k
  ;
  ADD %7, EAX         ; E = E + f
  MOV EBX, %3         ; EBX = A
  ROL EBX, 5          ; EBX = ROTATE_LEFT(A, 5)
  ADD %7, EBX         ; E = E + f + ROTATE_LEFT(A, 5)
  ADD %7, [w + 4*%1]  ; E = E + f + ROTATE_LEFT(A, 5) + w[i]
  ADD %7, K           ; E = E + f + ROTATE_LEFT(A, 5) + w[i] + k

  ;
  ; Actualizamos la variable 'B', que en la próxima ronda será
  ; 'C'.
  ;
  ; B = ROTATE_LEFT(B, 30)
  ;
  ROL %4, 30

%endmacro

;------------------------------------------------------------------------
section .data
  DEFAULT REL

  ;
  ; Vector w de 80 DWORDs para los 'w'. (Obs: TIMES es el
  ; equivalente a DUP de MASM)
  ;
  w TIMES 80 DD 0

;------------------------------------------------------------------------
section .text

get_sha1_asm:

  PUSH RBP
  MOV RBP, RSP

  ; Guardamos los registros de la convención C.
  PUSH RBX
  PUSH R12
  PUSH R13
  PUSH R14
  PUSH R15

  ;
  ; Obtenemos el desplazamiento final del último bloque que debe
  ; procesarse.
  ADD END_OF_BUFFER, BUFFER_PTR

  ;
  ; Inicializaciones.
  ;
  ; H[0] = 0x67452301;
  ; H[1] = 0xEFCDAB89;
  ; H[2] = 0x98BADCFE;
  ; H[3] = 0x10325476;
  ; H[4] = 0xC3D2E1F0;
  ;
  MOV DWORD [H0], 0x67452301
  MOV DWORD [H1], 0xEFCDAB89
  MOV DWORD [H2], 0x98BADCFE
  MOV DWORD [H3], 0x10325476
  MOV DWORD [H4], 0xC3D2E1F0

  ;
  ; A = H[0]
  ; B = H[1]
  ; C = H[2]
  ; D = H[3]
  ; E = H[4]
  ;
  MOV A, [H0]
  MOV B, [H1]
  MOV C, [H2]
  MOV D, [H3]
  MOV E, [H4]

loop:

  UPDATE_ABCDE 0,  BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 1,  BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 2,  BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 3,  BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 4,  BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 5,  BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 6,  BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 7,  BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 8,  BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 9,  BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 10, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 11, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 12, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 13, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 14, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 15, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 16, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 17, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 18, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 19, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 20, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 21, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 22, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 23, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 24, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 25, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 26, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 27, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 28, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 29, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 30, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 31, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 32, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 33, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 34, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 35, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 36, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 37, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 38, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 39, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 40, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 41, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 42, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 43, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 44, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 45, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 46, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 47, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 48, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 49, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 50, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 51, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 52, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 53, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 54, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 55, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 56, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 57, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 58, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 59, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 60, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 61, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 62, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 63, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 64, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 65, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 66, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 67, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 68, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 69, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 70, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 71, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 72, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 73, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 74, BUFFER_PTR, B, C, D, E, A
  UPDATE_ABCDE 75, BUFFER_PTR, A, B, C, D, E
  UPDATE_ABCDE 76, BUFFER_PTR, E, A, B, C, D
  UPDATE_ABCDE 77, BUFFER_PTR, D, E, A, B, C
  UPDATE_ABCDE 78, BUFFER_PTR, C, D, E, A, B
  UPDATE_ABCDE 79, BUFFER_PTR, B, C, D, E, A

  ;
  ; Actualizamos el hash:
  ;
  ; H[0] = H[0] + A;
  ; H[1] = H[1] + B;
  ; H[2] = H[2] + C;
  ; H[3] = H[3] + D;
  ; H[4] = H[4] + E;
  ;
  ADD A, [H0]
  ADD B, [H1]
  ADD C, [H2]
  ADD D, [H3]
  ADD E, [H4]

  MOV [H0], A
  MOV [H1], B
  MOV [H2], C
  MOV [H3], D
  MOV [H4], E

  ADD BUFFER_PTR, 64
  CMP BUFFER_PTR, END_OF_BUFFER
  JNE loop

  ;
  ;------------------------------------------------------------

  ;
  ; Restauramos los registros de la convención C.
  POP R15
  POP R14
  POP R13
  POP R12
  POP RBX
  POP RBP

  RET

;------------------------------------------------------------------------

