
;*****************************************************************
;
; SHA-1 sobre SIMD - Versión Ensamblador Vectorizado (SIMD)
;
; Pino, Jorge Daniel - JDanielPino@gmail.com
;
;*****************************************************************

; void get_sha1_simd(unsigned char* buffer, unsigned int size, unsigned int h[5]);
global get_sha1_simd
;
; Los parámetros de la función 'get_sha1_simd' se reciben en los
; siguientes registros:
;
; RDI:  'buffer'
; RSI:  'size'
; RDX:  'h[5]'
;

;----------------------------------------------------------------
;
; Definiciones de macros
;

;
; Asignamos aliases a los registros donde se reciben los
; parámetros, de manera tal que podamos referenciarlos con los
; nombres originales de esos parámetros. Obs: la macro 'xdefine',
; a diferencia de 'define', evita la redefinición.
%xdefine BUFFER_PTR    RDI
%xdefine END_OF_BUFFER RSI

;
; Vector h[5] donde se almacenará el hash calculado.
;
%xdefine H0 RDX
%xdefine H1 RDX + 4
%xdefine H2 RDX + 8
%xdefine H3 RDX + 12
%xdefine H4 RDX + 16

;
; Utilizamos a registros de propósito general para mantener las
; variables A, B, C, D y E del algoritmo de hashing.
;
%xdefine A R11D
%xdefine B R12D
%xdefine C R13D
%xdefine D R14D
%xdefine E R15D

;
; Buffer circular compuesto por 8 registros XMM que almacenan
; 32 elementos (4 por cada registro XMM) del vector w definido en
; el algoritmo. Este buffer se utiliza para el precálculo de los
; w[i] y seguidamente la suma w[i] + k, que es, en definitiva, el
; valor que realmente importa en el algoritmo (utilizado para el
; cálculo de la variable 'a'). La propiedad 'circular' se refiere
; a que el grupo de elementos calculados más allá de W7 se
; almacenan comenzando nuevamente por W0. Por ejemplo, los w
; correspondientes a la ronda 32 (que son w[32], w[33], w[34] y
; w[35]) se almacenan en W0, los otros cuatro de la ronda 36, en
; w[36], y así sucesivamente.
;
; Importante!
; A efectos de almacenar y recuperar los valores en el orden
; correcto (más precisamente, en el cálculo vectorizado 'w + k'),
; cada registro XMM0 almacena una fila de elementos w en el orden
; inverso (w[j-3], w[j-2], w[j-1], w[j]). Es fundamental tener en
; cuenta esto a la hora de leer los comentarios a lo largo de
; todo el código.
;
%xdefine W0 XMM0  ; w[i+3]  | w[i+2] | w[i+1] | w[i]
%xdefine W1 XMM1  ; w[i+7]  |       ...       | w[i+4]
%xdefine W2 XMM2  ; w[i+11] |       ...       | w[i+8]
%xdefine W3 XMM3  ; w[i+15] |       ...       | w[i+12]
%xdefine W4 XMM4  ; w[i+19] |       ...       | w[i+16]
%xdefine W5 XMM5  ; w[i+23] |       ...       | w[i+20]
%xdefine W6 XMM6  ; w[i+27] |       ...       | w[i+24]
%xdefine W7 XMM7  ; w[i+31] |       ...       | w[i+28]

;
; Usamos más registros XMM para almacenar las constantes K, las
; cuales se asignan en el código.
%xdefine K0 XMM8
%xdefine K1 XMM9
%xdefine K2 XMM10
%xdefine K3 XMM11

;
; Definimos los registros XMM que podemos usar libremente.
%xdefine XMM_AUX0 XMM12
%xdefine XMM_AUX1 XMM13
%xdefine XMM_AUX2 XMM14

;
; Usamos este registro XMM para almacenar el formato de
; reordenamiento en 'Little Endian' de los bytes leídos desde
; memoria.
%xdefine LITTLE_ENDIAN_BYTES_SHUFFLING XMM15

;
; Funciones F1, F2, F3 y F4 utilizadas en los cálculos
; intermedios del hash. Devuelven el resultado en EAX.
; Parámetros:
;
; %1: b
; %2: c
; %3: d
;
;
; F1 = (b & c) | ((~b) & d)
%macro F1 3
  MOV EAX, %2   ; EAX = c
  MOV EBX, %1   ; EBX = b
  AND EAX, EBX  ; EAX = (b & c)
  NOT EBX       ; EBX = ~b
  AND EBX, %3   ; EBX = ((~b) & d)
  OR  EAX, EBX  ; EAX = (b & c) | ((~b) & d)
%endmacro

;
; F2 = b ^ c ^ d;
%macro F2 3
  MOV EAX, %1  ; EAX = b
  XOR EAX, %2  ; EAX = b ^ c
  XOR EAX, %3  ; EAX = b ^ c ^d
%endmacro

;
; F3 = (b & c) | (b & d) | (c & d)
%macro F3 3
  MOV EAX, %1   ; EAX = b
  AND EAX, %2   ; EAX = b & c
  MOV EBX, %1   ; EBX = b
  AND EBX, %3   ; EBX = b & d
  OR  EAX, EBX  ; EAX = (b & c) | (b & d)
  MOV EBX, %2   ; EBX = c
  AND EBX, %3   ; EBX = c & d
  OR  EAX, EBX  ; EAX = (b & c) | (b & d) | (c & d)
%endmacro

;
; F4 = F2 = b ^ c ^ d;
%define F4 F2

;
; Esta macro actualiza los valores de las variables A, B, C, D y
; E llamando a la función F que corresponda, según el valor del
; índice 'i' (primer parámetro), que marca el número de ronda
; actual. Lo hace para dos rondas SHA1. Parámetros:
;
; %1: valor índice i de la ronda actual
; %2: A
; %3: B
; %4: C
; %5: D
; %6: E
;
; En la definición del algoritmo SHA1, las variables A, B, C, D y
; E se actualizan, después de realizar ciertos cálculos, moviendo
; circularmente los valores de dichas variables. De esta forma,
; E pasa a tener el valor de D, D el de C, y así sucesivamente.
; Podemos ahorrarnos muchas intrucciones de movimientos de datos
; si el llamador mismo se ocupa de rotar estas variables. Para
; ello, solo tiene que especificarlas en la lista de parámetros
; cuando invoca a esta macro, rotadas circularmente a la derecha
; dos veces (dado que la macro realiza la actualización para dos
; rondas). De esta forma, para la ronda 0, la llamada debería
; especificarse como:
;
;   UPDATE_ABCDE 0, A, B, C, D, E
;
; La siguiente actualización, en cambio, debería invocarse como:
;
;   UPDATE_ABCDE 2, D, E, A, B, C
;
; y así sucesivamente.
;
%macro UPDATE_ABCDE 6
    ;
    ; Rondas 0 a 19
    %if (%1 <= 19)
       %xdefine F F1
    ;
    ; Rondas 20 a 39
    %elif (%1 <= 39)
      %xdefine F F2
    ;
    ; Rondas 39 a 59
    %elif (%1 <= 59)
      %xdefine F F3
    ;
    ; Rondas 60 a 79
    %else
      %xdefine F F4
    %endif

    ;
    ; El resultado de F(B, C, D) lo obtenemos en EAX.
    F %3, %4, %5

    ;
    ; Actualizamos la variable 'E', que en la próxima ronda será
    ; 'A'.
    ;
    ; E = E + f + ROTATE_LEFT(A, 5) + w[i] + k
    ;
    MOV EBX, %2          ; EBX = A1
    ADD %6, EAX          ; E1  = E1 + f
    ROL EBX, 5           ; EBX = ROTATE_LEFT(A1, 5)
    ADD %6, [wk + 4*%1]  ; E1  = E1 + f + ROTATE_LEFT(A1, 5) + w[i] + k
    ADD %6, EBX          ; E1  = E1 + f + ROTATE_LEFT(A1, 5)

    ;
    ; Actualizamos la variable 'B1', que en la próxima ronda será
    ; 'C2'.
    ;
    ; B1 = ROTATE_LEFT(B1, 30)
    ;
    ROL %3, 30

    ;
    ; El código que sigue realiza la segunda actualización de A,
    ; B, C, D y E, especificando estas variables en el orden
    ; correcto según la rotación a derecha comentada en la
    ; descripción de la macro.
    ;
    MOV EBX, %6
    ROL EBX, 5
    ADD %5, [wk + 4*(%1+1)]  ; E2 = E1 + F(B, C , D) + w[i+1] + k
    ADD %5, EBX              ; E2 = E1 + F(B, C , D) + w[i+1] + k + ROTATE_LEFT(A, 5)
    F %2, %3, %4
    ADD %5, EAX              ; E2 = E1 + F(B, C , D)
    ROL %2, 30               ; C2 = ROTATE_LEFT(B1, 30)

%endmacro

;
; Esta macro centraliza las llamadas a las otras macros que
; calculan los elementos del vector WK de forma vectorizada.
%macro CALCULATE_WK 2

  %if (%1 < 16)
    CALCULATE_WK_FOR_ROUNDS_0_TO_15 %1, %2
  %elif (%1 < 32)
    CALCULATE_WK_FOR_ROUNDS_16_TO_31 %1
  %elif (%1 < 80)
    CALCULATE_WK_FOR_ROUNDS_32_TO_79 %1
  %endif

%endmacro

;
; Esta macro obtiene los primeros 16 'w' del algoritmo (de w[0] a
; w[15]), los almacena en el vector W0...W3 de XMMs, calcula los
; 'w[i] + k1' y los almacena en el vector 'wk' en memoria. Estos
; primeros 16 elementos se leen directamente del buffer y no
; requieren de mayor procesamiento. Parámetros:
;
; %1: número de ronda (solo válida para i = 0..15).
; %2: dirección del buffer.
;
; A los efectos de favorecer el mecanismo de 'pipelining' y la
; ejecución fuera de orden, la obtención de los cuatro elementos
; 'w' simultáneos usando SIMD se particiona en cuatro etapas,
; cada una de las cuales se distingue mediante el índice del
; número de ronda (primer parámetro). Esta partición en etapas
; favorece la intercalación con otras instrucciones
; independientes, y es precisamente esto lo que posibilita la
; ejecución fuera de orden.
;
%macro CALCULATE_WK_FOR_ROUNDS_0_TO_15 2

  ;
  ; Aplicamos una 'and' al índice 'i'. Esto nos permitirá
  ; distinguir el número de etapa.
  %assign i (%1 & 3)

  ;
  ; Etapa 1: rondas 0, 4, 8, 12. Simplemente leemos del buffer.
  ;
  %if (i == 0)

    ;
    ; Si es la primera ronda, empezar por el principio de vector
    ; armado con los registros XMM (W0, W1, ..., W7).
    %if (%1 == 0)
      %xdefine W W0
    %endif

    ;
    ; Leeremos de a bloques de 16 bytes en cada acceso. Los
    ; offsets que deberemos generar cada vez serán los
    ; siguientes: BUFFER_PTR (ronda 0), BUFFER_PTR + 16
    ; (ronda 4), BUFFER_PTR + 32 (ronda 8) y BUFFER_PTR + 48
    ; (ronda 12). Eso es lo que hace la expresión debajo.
    MOVDQU W, [%2 + 4*%1]

  ;
  ; Etapa 2: rondas 1, 5, 9, 13. Ejecutamos el formateo de los
  ; datos.
  ;
  %elif (i == 1)
    PSHUFB W, LITTLE_ENDIAN_BYTES_SHUFFLING
  ;
  ; Etapa 3: rondas 2, 6, 10, 14. Calculamos w + k y guardamos el
  ; resultado en el vector 'wk' en memoria.
  %elif (i == 2)
    MOVDQU XMM_AUX0, W

  ;
  ; Etapa 4: rondas 3, 7, 11, 15.
  ;
  %elif (i == 3)

    PADDD  XMM_AUX0, K0

    ;
    ; Debemos guardar los 'w + k' obtenidos en el vector 'wk'.
    ; Guardaremos los primeros cuatro wk en 'wk + 0' (ronda 3),
    ; los siguientes cuatro en 'wk + 16' (ronda 7), los
    ; siguientes en 'wk + 32' (ronda 11), y los últimos en
    ; 'wk + 48' (ronda 15). Eso es lo que hace la expresión
    ; debajo, utilizando como índice al número de ronda (primer
    ; parámetro: %1).
    MOVDQU [wk + 4*(%1 - 3)], XMM_AUX0

    ;
    ; Reasignamos el símbolo W para que apunte al siguiente
    ; registro XMM en el vector de w's que se está armando.
    %if (%1 == 3)
      %xdefine W W1
    %elif (%1 == 7)
      %xdefine W W2
    %elif (%1 == 11)
      %xdefine W W3
    %endif

  %endif

%endmacro

;
; Esta macro obtiene los elementos 'w[16]' a 'w[31] del
; algoritmo, los almacena en el vector W4...W7 de XMMs, calcula
; los 'w[i] + k' y los almacena en el vector 'wk' en memoria.
; Lo hace de forma vectorizada calculando de a 4 elementos a la
; vez. Con esta macro se inicia el mecanismo de 'pipelining'
; combinado con el método de optimización 'loop unrolling', en
; el cual un bucle se depliega para ahorrar las intrucciones de
; control del mismo.
;
; Parámetro:
;
; %1: dirección del buffer con los datos.
;
; A los efectos de favorecer el mecanismo de 'pipelining' y la
; ejecución fuera de orden, la obtención de los cuatro elementos
; 'w' simultáneos usando SIMD se particiona en cuatro etapas,
; cada una de las cuales se distingue mediante el índice del
; número de ronda (primer parámetro). Esta partición en etapas
; favorece la intercalación con otras instrucciones
; independientes, y es precisamente esto lo que posibilita la
; ejecución fuera de orden.
;
%macro CALCULATE_WK_FOR_ROUNDS_16_TO_31 1

  ;
  ; El cálculo del elemento w[i] requieren de los elementos
  ; w[i-16], w[i-14], w[i-8] y w[i-3]. De esta forma, el elemento
  ; w[16] depende de w[0], w[2], w[8] y w[13]; el w[17] de w[1],
  ; w[3], w[9] y w[14], y así sucesivamente. Puesto que los
  ; elementos del vector w se almacenan en registros XMM, la
  ; única forma que tenemos de acceder y navegar los distintos w
  ; conforme avanzamos en el cálculo de los w[i], es redefiniendo
  ; un conjunto de macros cada vez. A estas macros las
  ; denominamos:
  ;
  ;   W_MINUS_16, para los elementos w[i-16], w[i-15], w[i-14] y
  ;   w[i-13];
  ;
  ;   W_MINUS_12, para los elementos w[i-12], w[i-11], w[i-10] y
  ;   w[i-9];
  ;
  ;   W_MINUS_8, para los elementos w[i-8], w[i-7], w[i-6] y
  ;   w[i-5];
  ;
  ;   W_MINUS_4, para los elementos w[i-4], w[i-3], w[i-2] y
  ;   w[i-1];
  ;
  ; El código que sigue incluye las redefiniciones de estas
  ; macros con el objeto de ir rellenando correctamente al vector
  ; W.
  ;

  ;
  ; Aplicamos una 'and' al índice 'i'. Esto nos permitirá
  ; distinguir el número de etapa.
  %assign i (%1 & 3)

  ;
  ; Etapa 1: rondas 16, 20, 24, 28.
  ;
  %if (i == 0)

    %if (%1 == 16)
      %xdefine W_MINUS_16 W0
      %xdefine W_MINUS_12 W1
      %xdefine W_MINUS_8  W2
      %xdefine W_MINUS_4  W3
      %xdefine W W4
      %xdefine K K0
    %endif

    MOVDQU W, W_MINUS_16             ;        W = w[i-13] | w[i-14] | w[i-15] | w[i-16]
    MOVDQU XMM_AUX0, W_MINUS_12      ; XMM_AUX0 = w[i-9]  | w[i-10] | w[i-11] | w[i-12]
    PALIGNR XMM_AUX0, W_MINUS_16, 8  ; XMM_AUX0 = w[i-11] | w[i-12] | w[i-13] | w[i-14]
    MOVDQU XMM_AUX1, W_MINUS_4       ; XMM_AUX1 = w[i-1]  | w[i-2]  | w[i-3]  | w[i-4]
    PSRLDQ XMM_AUX1, 4               ; XMM_AUX1 =    0    | w[i-1]  | w[i-2]  | w[i-3]

  ;
  ; Etapa 2: rondas 17, 21, 25, 29.
  ;
  %elif (i == 1)
    ;
    ; Con todos los registros cargados, ejecutamos la cadena de XORs.
    PXOR W, XMM_AUX0 ; w[i-13] XOR w[i-11]                  | ... | w[i-16] XOR w[i-14]
    PXOR W, W_MINUS_8; w[i-13] XOR w[i-11] XOR w[i-5]       | ... | w[i-16] XOR w[i-14] XOR w[i-8]
    PXOR W, XMM_AUX1 ; w[i-13] XOR w[i-11] XOR w[i-5] XOR 0 | ... | w[i-16] XOR w[i-14] XOR w[i-8] XOR w[i-3]

    ;
    ; Realizamos las rotaciones sobre los elementos de la
    ; vectorización. Dado que no existe una instrucción para
    ; realizar rotaciones en DWORDs empaquetados en registros XMMs,
    ; podemos emularla haciendo una copia del vector, rotando n
    ; bits en una dirección al original, y (32 - n) bits a la copia
    ; hacia el otro lado. Luego, realizar un POR entre el original
    ; y la copia.
    MOVDQU XMM_AUX0, W
    PSLLD  W, 1
    PSRLD  XMM_AUX0, 31
    POR    W, XMM_AUX0  ; W = w'[i+3] | w[i+2] | w[i+1] | w[i]  -Obs: w'[i+3]: cálculo parcial de w[i+3]

  ;
  ; Etapa 3: rondas 18, 22, 26, 30.
  ;
  %elif (i == 2)
    ;
    ; Ahora, necesitamos completar las operaciones sobre el cuarto
    ; elemento de la vectorización (w[i+3]), que dependía del
    ; primero (w[i]). Esto es posible gracias a la propiedad:
    ;
    ;   Rotate_Left(A XOR B, n) = Rotate_Left(A, n) XOR Rotate_Left(B, n)
    ;
    ; Esta propiedad permite que completemos el cálculo sobre w'[19]
    ; a posteriori, aplicándole la rotación a una copia de w[16], y
    ; luego ejecutando el XOR final.
    ;
    MOVDQU XMM_AUX0, W    ; XMM_AUX0 = w'[i+3] | w[i+2] | w[i+1] | w[i]
    MOVDQU XMM_AUX1, W    ; XMM_AUX1 = w'[i+3] | w[i+2] | w[i+1] | w[i]
    PSRLD  XMM_AUX0, 31
    PSLLD  XMM_AUX1, 1

    POR    XMM_AUX0, XMM_AUX1  ; XMM_AUX0 = ... | ... | ... | Rotate_Left(w[i], 1)
    PSLLDQ XMM_AUX0, 12        ; XMM_AUX0 = Rotate_Left(w[i], 1) |    0   |    0   |  0
    PXOR   W, XMM_AUX0         ;        W =       w[i+3]         | w[i+2] | w[i+1] | w[i]

  ;
  ; Etapa 4: rondas 19, 23, 27, 31.
  ;
  %elif (i == 3)
    ;
    ; Calculamos la suma 'w[i] + k', y la guardamos en memoria.
    MOVDQU XMM_AUX0, W  ; XMM_AUX0 = w[i+3]     | w[i+2]     | w[i+1]      | w[i]
    PADDD  XMM_AUX0, K  ; XMM_AUX0 = w[i+3] + k | w[i+2] + k | w[i+1]  + k | w[i] + k

    ;
    ; Debemos guardar los 'w + k' obtenidos en el vector 'wk'.
    ; Guardaremos los primeros cuatro wk en 'wk + 64' (ronda 19),
    ; los siguientes cuatro en 'wk + 80' (ronda 23), los
    ; siguientes en 'wk + 96' (ronda 27), y los últimos en
    ; 'wk + 112' (ronda 31). Eso es lo que hace la expresión
    ; debajo, utilizando como índice al número de ronda (primer
    ; parámetro: %1).
    MOVDQU [wk + 64 + 4*(%1 - 19)], XMM_AUX0

    ;
    ; Avanzamos en el armado del vector W. Conforme vamos
    ; calculando de a cuatro elementos de W, nos movemos al
    ; próximo registro, reasignándolo.
    ;
    %if (%1 == 19)
      %xdefine W_MINUS_16 W1
      %xdefine W_MINUS_12 W2
      %xdefine W_MINUS_8  W3
      %xdefine W_MINUS_4  W4
      %xdefine W W5
      %xdefine K K1

    %elif (%1 == 23)
      %xdefine W_MINUS_16 W2
      %xdefine W_MINUS_12 W3
      %xdefine W_MINUS_8  W4
      %xdefine W_MINUS_4  W5
      %xdefine W W6

    %elif (%1 == 27)
      %xdefine W_MINUS_16 W3
      %xdefine W_MINUS_12 W4
      %xdefine W_MINUS_8  W5
      %xdefine W_MINUS_4  W6
      %xdefine W W7
    %endif

  %endif

%endmacro

;
; Esta macro obtiene los elementos finales 'w[32]' a 'w[79]' del
; algoritmo. Los valores w calculados reemplazan a los
; existentes, comenzando por el w[0]. De esta forma, los valores
; w[32], w[33], w[34] y w[35], reemplazan a los antiguos valores
; w[0], w[1], w[2] y w[3] del vector W0, y así sucesivamente. Es
; por ello que el buffer W0..W7 se considera circular. Esto es
; posible porque el rango de las dependencias de los w lo
; permiten (es decir, no pisamos ningún w del cual dependa algún
; otro). Además, en última instancia, lo que realmente interesa
; en el algoritmo es el valor de los 'w[i] + k'. Para esta última
; etapa en el procesamiento del bloque de datos de 64 bytes, el
; 'pipelining' de 4 elementos es completo, ya que cualquier w
; depende de otros w anteriores que fueron completamente
; calculados.
;
; Parámetro:
;
; %1: dirección del buffer con los datos.
;
; A los efectos de favorecer el mecanismo de 'pipelining' y la
; ejecución fuera de orden, la obtención de los cuatro elementos
; 'w' simultáneos usando SIMD se particiona en cuatro etapas,
; cada una de las cuales se distingue mediante el índice del
; número de ronda (primer parámetro). Esta partición en etapas
; favorece la intercalación con otras instrucciones
; independientes, y es precisamente esto lo que posibilita la
; ejecución fuera de orden.
;
%macro CALCULATE_WK_FOR_ROUNDS_32_TO_79 1

  ;
  ; El cálculo del elemento w[i] requieren de los elementos
  ; w[i-16], w[i-14], w[i-8] y w[i-3]. De esta forma, el elemento
  ; w[16] depende de w[0], w[2], w[8] y w[13]; el w[17] de w[1],
  ; w[3], w[9] y w[14], y así sucesivamente. Puesto que los
  ; elementos del vector w se almacenan en registros XMM, la
  ; única forma que tenemos de acceder y navegar los distintos w
  ; conforme avanzamos en el cálculo de los w[i], es redefiniendo
  ; un conjunto de macros cada vez. A estas macros las
  ; denominamos:
  ;
  ;   W_MINUS_32, para los elementos w[i-32], w[i-31], w[i-30] y
  ;   w[i-29];
  ;
  ;   W_MINUS_28, para los elementos w[i-28], w[i-27], w[i-26] y
  ;   w[i-25];
  ;
  ;   W_MINUS_16, para los elementos w[i-16], w[i-15], w[i-14] y
  ;   w[i-13];
  ;
  ;   W_MINUS_8, para los elementos w[i-8], w[i-7], w[i-6] y
  ;   w[i-5];
  ;
  ; El código que sigue incluye las redefiniciones de estas
  ; macros con el objeto de ir rellenando correctamente al vector
  ; W.
  ;

  ;
  ; Aplicamos una 'and' al índice 'i'. Esto nos permitirá
  ; distinguir el número de etapa.
  %assign i (%1 & 3)

  ;
  ; Etapa 1: rondas 32, 36, 40, 44, 48, 52, 56, 60, 64, 68, 72, 76.
  ;
  %if (i == 0)

    %if (%1 == 32)
      %xdefine W W0
      %xdefine W_MINUS_28 W1
      %xdefine W_MINUS_16 W4
      %xdefine W_MINUS_8  W6
      %xdefine W_MINUS_4  W7
    %endif
                        ; W = w[i-29] | w[i-30] | w[i-31] | w[i-32]
    PXOR W, W_MINUS_28  ; W = w[i-29] XOR w[i-25]             | ... | w[i-32] XOR w[i-28]
    PXOR W, W_MINUS_16  ; W = w[i-29] XOR w[i-25] XOR w[i-13] | ... | w[i-32] XOR w[i-28] XOR w[i-16]

  ;
  ; Etapa 2: rondas 33, 37, 41, 45, 49, 53, 57, 61, 65, 69, 73, 77.
  ;
  %elif (i == 1)
    MOVDQU  XMM_AUX0, W_MINUS_4     ; XMM_AUX0 = w[i-1] | w[i-2] | w[i-3] | w[i-4]
    PALIGNR XMM_AUX0, W_MINUS_8, 8  ; XMM_AUX0 = w[i-3] | w[i-4] | w[i-5] | w[i-6]

    PXOR W, XMM_AUX0  ; W = ... |w[i-32] XOR w[i-28] XOR w[i-16] XOR w[i-6]

  ;
  ; Etapa 3: rondas 34, 38, 42, 46, 50, 54, 58, 62, 66, 70, 74, 78.
  ;
  %elif (i == 2)
    ;
    ; Realizamos las rotaciones sobre los elementos de la
    ; vectorización. Dado que no existe una instrucción para
    ; realizar rotaciones en DWORDs empaquetados en registros XMMs,
    ; podemos emularla haciendo una copia del vector, rotando n
    ; bits en una dirección al original, y (32 - n) bits a la copia
    ; hacia el otro lado. Luego, realizar un POR entre el original
    ; y la copia.
    MOVDQU XMM_AUX0, W
    PSLLD  W, 2
    PSRLD  XMM_AUX0, 30
    POR    W, XMM_AUX0  ; W = w[i] | w[i+1] | w[i+2] | w[i+3]

  ;
  ; Etapa 4: rondas 35, 39, 43, 47, 51, 55, 59, 63, 67, 71, 75, 79.
  ;
  %elif (i == 3)
    ;
    ; Calculamos la suma 'w[i] + k', y la guardamos en memoria.
    MOVDQU XMM_AUX0, W  ; XMM_AUX0 = w[i]     | w[i+1]     | w[i+2]      | w[i+3]
    PADDD  XMM_AUX0, K  ; XMM_AUX0 = w[i] + k | w[i+1] + k | w[i+2]  + k | w[i+3] + k

    ;
    ; Debemos guardar los 'w + k' obtenidos en el vector 'wk'.
    ; Guardaremos los primeros cuatro wk en 'wk + 128' (ronda
    ; 35), los siguientes cuatro en 'wk + 144' (ronda 39), los
    ; siguientes en 'wk + 160' (ronda 43),y así sucesivamente.
    ; Eso es lo que hace la expresión debajo, utilizando como
    ; índice al número de ronda (primer parámetro: %1).
    MOVDQU [wk + 128 + 4*(%1 - 35)], XMM_AUX0

    %if (%1 == 35 || %1 == 67)
      %xdefine W W1
      %xdefine W_MINUS_28 W2
      %xdefine W_MINUS_16 W5
      %xdefine W_MINUS_8  W7
      %xdefine W_MINUS_4  W0

    %elif (%1 == 39 || %1 == 71)
      %xdefine W W2
      %xdefine W_MINUS_28 W3
      %xdefine W_MINUS_16 W6
      %xdefine W_MINUS_8  W0
      %xdefine W_MINUS_4  W1

      %if (%1 == 39)
	%xdefine K K2
      %endif

    %elif (%1 == 43 || %1 == 75)
      %xdefine W W3
      %xdefine W_MINUS_28 W4
      %xdefine W_MINUS_16 W7
      %xdefine W_MINUS_8  W1
      %xdefine W_MINUS_4  W2

    %elif (%1 == 47 || %1 == 79)
      %xdefine W W4
      %xdefine W_MINUS_28 W5
      %xdefine W_MINUS_16 W0
      %xdefine W_MINUS_8  W2
      %xdefine W_MINUS_4  W3

    %elif (%1 == 51)
      %xdefine W W5
      %xdefine W_MINUS_28 W6
      %xdefine W_MINUS_16 W1
      %xdefine W_MINUS_8  W3
      %xdefine W_MINUS_4  W4

    %elif (%1 == 55)
      %xdefine W W6
      %xdefine W_MINUS_28 W7
      %xdefine W_MINUS_16 W2
      %xdefine W_MINUS_8  W4
      %xdefine W_MINUS_4  W5

    %elif (%1 == 59)
      %xdefine W W7
      %xdefine W_MINUS_28 W0
      %xdefine W_MINUS_16 W3
      %xdefine W_MINUS_8  W5
      %xdefine W_MINUS_4  W6
      %xdefine K K3

    %elif (%1 == 63)
      %xdefine W W0
      %xdefine W_MINUS_28 W1
      %xdefine W_MINUS_16 W4
      %xdefine W_MINUS_8  W6
      %xdefine W_MINUS_4  W7
    %endif
  %endif

%endmacro

;------------------------------------------------------------------------
section .data align=16
  DEFAULT REL

  ;
  ; En el vector 'wk' almacenaremos los 80 'w[i] + k'
  ; correspondientes a las 80 rondas del algoritmo para un
  ; bloque. Los valores acá almacenados se utilizan en el
  ; cálculo:
  ;
  ;    temp = ROTATE_LEFT(a, 5) + f + e + (k + w[j])
  ;        ...
  ;    a = temp
  ;
  ; Para optimizar la suma, ya cargamos todos los 'k' en el vector
  ; 'wk'.
  ;
  wk TIMES 80 DD 0x0

  ;
  ; Cada secuencia de 16 bytes que se lee desde el buffer será
  ; cargada en los registros XMM en el mismo orden en que
  ; encuentran en dicho buffer. Esos bloques de 16 bytes en
  ; realidad se interpretan como 4 palabras DWORD. Por lo tanto,
  ; habrá que reorganizar los datos en formato 'Little Endian'
  ; antes de operar con ellos según lo indica el algoritmo. Para
  ; ello, podemos utilizar la instrucción PSHUFB, de la extensión
  ; SSE3. Esta instrucción utiliza índices que determinan dónde
  ; debe reubicarse cada uno de los 16 bytes de un registro XMM.
  ; La variable que sigue define el orden de reubicación para
  ; formatear correctamente los 4 DWORDs.
  little_endian_bytes_shuffling_values DB 0x03, 0x02, 0x01, 0x00,\
                                          0x07, 0x06, 0x05, 0x04,\
                                          0x0b, 0x0a, 0x09, 0x08,\
                                          0x0f, 0x0e, 0x0d, 0x0c
  ;
  ; Constantes 'k' utilizadas para el cálculo de la variable 'a'
  ; en la expresión:
  ;
  ;    temp = ROTATE_LEFT(a, 5) + f + e + k + w[j]
  ;        ...
  ;    a = temp
  ;
  ; La razón por la que estas variables aparecen cuadruplicadas
  ; es para sumarlas de forma vectorizada sobre los registros
  ; XMM.
  k1 DD 0x5A827999, 0x5A827999, 0x5A827999, 0x5A827999
  k2 DD 0x6ED9EBA1, 0x6ED9EBA1, 0x6ED9EBA1, 0x6ED9EBA1
  k3 DD 0x8F1BBCDC, 0x8F1BBCDC, 0x8F1BBCDC, 0x8F1BBCDC
  k4 DD 0xCA62C1D6, 0xCA62C1D6, 0xCA62C1D6, 0xCA62C1D6


;------------------------------------------------------------------------
section .text

get_sha1_simd:

  PUSH RBP
  MOV RBP, RSP

  ; Guardamos los registros de la convención C.
  PUSH RBX
  PUSH R12
  PUSH R13
  PUSH R14
  PUSH R15

  ;
  ; Obtenemos el desplazamiento final del último bloque que debe
  ; procesarse.
  ADD END_OF_BUFFER, BUFFER_PTR

  ;
  ; Inicializamos los valores para el formateo en 'Little Endian'
  ; de los datos leídos del buffer en un registro XMM.
  MOVDQU LITTLE_ENDIAN_BYTES_SHUFFLING, [little_endian_bytes_shuffling_values]

  ;
  ; Inicializaciones.
  ;
  ; H[0] = 0x67452301;
  ; H[1] = 0xEFCDAB89;
  ; H[2] = 0x98BADCFE;
  ; H[3] = 0x10325476;
  ; H[4] = 0xC3D2E1F0;
  ;
  MOV DWORD [H0], 0x67452301
  MOV DWORD [H1], 0xEFCDAB89
  MOV DWORD [H2], 0x98BADCFE
  MOV DWORD [H3], 0x10325476
  MOV DWORD [H4], 0xC3D2E1F0

  ;
  ; Más inicializaciones.
  ;
  ; K0 = 0x5A827999 | 0x5A827999 | 0x5A827999 | 0x5A827999
  MOVDQU K0, [k1]

  ; K1 = 0x6ED9EBA1 | 0x6ED9EBA1 | 0x6ED9EBA1 | 0x6ED9EBA1
  MOVDQU K1, [k2]

  ; K2 = 0x8F1BBCDC | 0x8F1BBCDC | 0x8F1BBCDC | 0x8F1BBCDC
  MOVDQU K2, [k3]

  ; K3 = 0xCA62C1D6 | 0xCA62C1D6 | 0xCA62C1D6 | 0xCA62C1D6
  MOVDQU K3, [k4]

  ;
  ; Calculamos por adelantado 16 elementos del vector W (64
  ; bytes).
  CALCULATE_WK 0, BUFFER_PTR
  CALCULATE_WK 1, BUFFER_PTR
  CALCULATE_WK 2, BUFFER_PTR
  CALCULATE_WK 3, BUFFER_PTR  ; Obtenemos W0: w[0] | w[1] | w[2] | w[3]

  CALCULATE_WK 4, BUFFER_PTR
  CALCULATE_WK 5, BUFFER_PTR
  CALCULATE_WK 6, BUFFER_PTR
  CALCULATE_WK 7, BUFFER_PTR  ; Obtenemos W1: w[4] | w[5] | w[6] | w[7]

  CALCULATE_WK 8,  BUFFER_PTR
  CALCULATE_WK 9,  BUFFER_PTR
  CALCULATE_WK 10, BUFFER_PTR
  CALCULATE_WK 11, BUFFER_PTR  ; Obtenemos W2: w[8] | w[9] | w[10] | w[11]

  CALCULATE_WK 12, BUFFER_PTR
  CALCULATE_WK 13, BUFFER_PTR
  CALCULATE_WK 14, BUFFER_PTR
  CALCULATE_WK 15, BUFFER_PTR  ; Obtenemos W3: w[12] | w[13] | w[14] | w[15]

  ;
  ; A = H[0]
  ; B = H[1]
  ; C = H[2]
  ; D = H[3]
  ; E = H[4]
  ;
  MOV A, [H0]
  MOV B, [H1]
  MOV C, [H2]
  MOV D, [H3]
  MOV E, [H4]

;
; Ciclo del algoritmo desplegado (loop unrolling).
;
loop:

  CALCULATE_WK 16, BUFFER_PTR
  UPDATE_ABCDE 0,  A, B, C, D, E
  CALCULATE_WK 17, BUFFER_PTR

  CALCULATE_WK 18, BUFFER_PTR
  UPDATE_ABCDE 2,  D, E, A, B, C
  CALCULATE_WK 19, BUFFER_PTR     ; Obtenemos w[16], w[17], w[18] y w[19].

  CALCULATE_WK 20, BUFFER_PTR
  UPDATE_ABCDE 4,  B, C, D, E, A
  CALCULATE_WK 21, BUFFER_PTR

  CALCULATE_WK 22, BUFFER_PTR
  UPDATE_ABCDE 6,  E, A, B, C, D
  CALCULATE_WK 23, BUFFER_PTR     ; Obtenemos w[20], w[21], w[22] y w[23].

  CALCULATE_WK 24, BUFFER_PTR
  UPDATE_ABCDE 8,  C, D, E, A, B
  CALCULATE_WK 25, BUFFER_PTR

  CALCULATE_WK 26, BUFFER_PTR
  UPDATE_ABCDE 10, A, B, C, D, E
  CALCULATE_WK 27, BUFFER_PTR     ; Obtenemos w[24], w[25], w[26] y w[27].

  CALCULATE_WK 28, BUFFER_PTR
  UPDATE_ABCDE 12, D, E, A, B, C
  CALCULATE_WK 29, BUFFER_PTR

  CALCULATE_WK 30, BUFFER_PTR
  UPDATE_ABCDE 14, B, C, D, E, A
  CALCULATE_WK 31, BUFFER_PTR     ; Obtenemos w[28], w[29], w[30] y w[31].

  CALCULATE_WK 32, BUFFER_PTR
  UPDATE_ABCDE 16, E, A, B, C, D
  CALCULATE_WK 33, BUFFER_PTR

  CALCULATE_WK 34, BUFFER_PTR
  UPDATE_ABCDE 18, C, D, E, A, B
  CALCULATE_WK 35, BUFFER_PTR     ; Obtenemos w[32], w[33], w[34] y w[35].

  CALCULATE_WK 36, BUFFER_PTR
  UPDATE_ABCDE 20, A, B, C, D, E
  CALCULATE_WK 37, BUFFER_PTR

  CALCULATE_WK 38, BUFFER_PTR
  UPDATE_ABCDE 22, D, E, A, B, C
  CALCULATE_WK 39, BUFFER_PTR     ; Obtenemos w[36], w[37], w[38] y w[39].

  CALCULATE_WK 40, BUFFER_PTR
  UPDATE_ABCDE 24, B, C, D, E, A
  CALCULATE_WK 41, BUFFER_PTR

  CALCULATE_WK 42, BUFFER_PTR
  UPDATE_ABCDE 26, E, A, B, C, D
  CALCULATE_WK 43, BUFFER_PTR     ; Obtenemos w[40], w[41], w[42] y w[43].

  CALCULATE_WK 44, BUFFER_PTR
  UPDATE_ABCDE 28, C, D, E, A, B
  CALCULATE_WK 45, BUFFER_PTR

  CALCULATE_WK 46, BUFFER_PTR
  UPDATE_ABCDE 30, A, B, C, D, E
  CALCULATE_WK 47, BUFFER_PTR     ; Obtenemos w[44], w[45], w[46] y w[47].

  CALCULATE_WK 48, BUFFER_PTR
  UPDATE_ABCDE 32, D, E, A, B, C
  CALCULATE_WK 49, BUFFER_PTR

  CALCULATE_WK 50, BUFFER_PTR
  UPDATE_ABCDE 34, B, C, D, E, A
  CALCULATE_WK 51, BUFFER_PTR     ; Obtenemos w[48], w[49], w[50] y w[51].

  CALCULATE_WK 52, BUFFER_PTR
  UPDATE_ABCDE 36, E, A, B, C, D
  CALCULATE_WK 53, BUFFER_PTR

  CALCULATE_WK 54, BUFFER_PTR
  UPDATE_ABCDE 38, C, D, E, A, B
  CALCULATE_WK 55, BUFFER_PTR     ; Obtenemos w[52], w[53], w[54] y w[55].

  CALCULATE_WK 56, BUFFER_PTR
  UPDATE_ABCDE 40, A, B, C, D, E
  CALCULATE_WK 57, BUFFER_PTR

  CALCULATE_WK 58, BUFFER_PTR
  UPDATE_ABCDE 42, D, E, A, B, C
  CALCULATE_WK 59, BUFFER_PTR     ; Obtenemos w[56], w[57], w[58] y w[59].

  CALCULATE_WK 60, BUFFER_PTR
  UPDATE_ABCDE 44, B, C, D, E, A
  CALCULATE_WK 61, BUFFER_PTR

  CALCULATE_WK 62, BUFFER_PTR
  UPDATE_ABCDE 46, E, A, B, C, D
  CALCULATE_WK 63, BUFFER_PTR     ; Obtenemos w[60], w[61], w[62] y w[63].

  CALCULATE_WK 64, BUFFER_PTR
  UPDATE_ABCDE 48, C, D, E, A, B
  CALCULATE_WK 65, BUFFER_PTR

  CALCULATE_WK 66, BUFFER_PTR
  UPDATE_ABCDE 50, A, B, C, D, E
  CALCULATE_WK 67, BUFFER_PTR     ; Obtenemos w[64], w[65], w[66] y w[67].

  CALCULATE_WK 68, BUFFER_PTR
  UPDATE_ABCDE 52, D, E, A, B, C
  CALCULATE_WK 69, BUFFER_PTR

  CALCULATE_WK 70, BUFFER_PTR
  UPDATE_ABCDE 54, B, C, D, E, A
  CALCULATE_WK 71, BUFFER_PTR     ; Obtenemos w[68], w[69], w[70] y w[71].

  CALCULATE_WK 72, BUFFER_PTR
  UPDATE_ABCDE 56, E, A, B, C, D
  CALCULATE_WK 73, BUFFER_PTR

  CALCULATE_WK 74, BUFFER_PTR
  UPDATE_ABCDE 58, C, D, E, A, B
  CALCULATE_WK 75, BUFFER_PTR     ; Obtenemos w[72], w[73], w[74] y w[75].

  CALCULATE_WK 76, BUFFER_PTR
  UPDATE_ABCDE 60, A, B, C, D, E
  CALCULATE_WK 77, BUFFER_PTR

  CALCULATE_WK 78, BUFFER_PTR
  UPDATE_ABCDE 62, D, E, A, B, C
  CALCULATE_WK 79, BUFFER_PTR     ; Obtenemos w[76], w[77], w[78] y w[79].

  ;
  ; Nos deplazamos al siguiente bloque del buffer sumando 64
  ; bytes al puntero. Si este era el último, saltamos a la
  ; actualización de las variables A, B, C, D y E, sin calcular
  ; por adelantado ningún otro elemento del vector W.
  ADD BUFFER_PTR, 64
  CMP BUFFER_PTR, END_OF_BUFFER
  JE last_abcde_updates

  ;
  ; Comienza el cálculo por adelantado de los próximos 16
  ; elementos del vector W y los w + k, que se utilizarán para
  ; actualizar los valores de las variables A, B, C, D y E en la
  ; próxima iteración.
  ;
  CALCULATE_WK 0,  BUFFER_PTR
  UPDATE_ABCDE 64, B, C, D, E, A
  CALCULATE_WK 1,  BUFFER_PTR

  CALCULATE_WK 2,  BUFFER_PTR
  UPDATE_ABCDE 66, E, A, B, C, D
  CALCULATE_WK 3,  BUFFER_PTR     ; Obtenemos w[0], w[1], w[2] y w[3].

  CALCULATE_WK 4,  BUFFER_PTR
  UPDATE_ABCDE 68, C, D, E, A, B
  CALCULATE_WK 5,  BUFFER_PTR

  CALCULATE_WK 6,  BUFFER_PTR
  UPDATE_ABCDE 70, A, B, C, D, E
  CALCULATE_WK 7,  BUFFER_PTR     ; Obtenemos w[4], w[5], w[6] y w[7].

  CALCULATE_WK 8,  BUFFER_PTR
  UPDATE_ABCDE 72, D, E, A, B, C
  CALCULATE_WK 9,  BUFFER_PTR

  CALCULATE_WK 10, BUFFER_PTR
  UPDATE_ABCDE 74, B, C, D, E, A
  CALCULATE_WK 11, BUFFER_PTR     ; Obtenemos w[8], w[9], w[10] y w[11].

  CALCULATE_WK 12, BUFFER_PTR
  UPDATE_ABCDE 76, E, A, B, C, D
  CALCULATE_WK 13, BUFFER_PTR

  CALCULATE_WK 14, BUFFER_PTR
  UPDATE_ABCDE 78, C, D, E, A, B
  CALCULATE_WK 15, BUFFER_PTR     ; Obtenemos w[12], w[13], w[14] y w[15].

  ;
  ; Actualizamos el hash:
  ;
  ; H[0] = H[0] + A;
  ; H[1] = H[1] + B;
  ; H[2] = H[2] + C;
  ; H[3] = H[3] + D;
  ; H[4] = H[4] + E;
  ;
  ADD A, [H0]
  ADD B, [H1]
  ADD C, [H2]
  ADD D, [H3]
  ADD E, [H4]

  MOV [H0], A
  MOV [H1], B
  MOV [H2], C
  MOV [H3], D
  MOV [H4], E

  ;
  ; Iteramos.
  JMP loop
  ;
  ;------------------------------------------------------------

;
; Última actualización.
;
last_abcde_updates:
  UPDATE_ABCDE 64, B, C, D, E, A
  UPDATE_ABCDE 66, E, A, B, C, D
  UPDATE_ABCDE 68, C, D, E, A, B
  UPDATE_ABCDE 70, A, B, C, D, E
  UPDATE_ABCDE 72, D, E, A, B, C
  UPDATE_ABCDE 74, B, C, D, E, A
  UPDATE_ABCDE 76, E, A, B, C, D
  UPDATE_ABCDE 78, C, D, E, A, B

  ;
  ; Actualizamos por última vez el hash:
  ;
  ; H[0] = H[0] + A;
  ; H[1] = H[1] + B;
  ; H[2] = H[2] + C;
  ; H[3] = H[3] + D;
  ; H[4] = H[4] + E;
  ;
  ADD A, [H0]
  ADD B, [H1]
  ADD C, [H2]
  ADD D, [H3]
  ADD E, [H4]

  MOV [H0], A
  MOV [H1], B
  MOV [H2], C
  MOV [H3], D
  MOV [H4], E
  ;
  ;------------------------------------------------------------

  ;
  ; Restauramos los registros de la convención C.
  POP R15
  POP R14
  POP R13
  POP R12
  POP RBX
  POP RBP

  RET

;------------------------------------------------------------------------

