
;*****************************************************************
;
; SHA-256 sobre SIMD - Versión Ensamblador Escalar.
;
; Pino, Jorge Daniel - JDanielPino@gmail.com
;
;*****************************************************************

; void get_sha256_asm(unsigned char* buffer, unsigned int size, unsigned int h[8]);
global get_sha256_asm
;
; Los parámetros de la función 'get_sha256_asm' se reciben en los
; siguientes registros:
;
; RDI:  'buffer'
; RSI:  'size'
; RDX:  'h[8]'
;

;----------------------------------------------------------------
;
; Definiciones de macros
;

;
; Asignamos aliases a los registros donde se reciben los
; parámetros, de manera tal que podamos referenciarlos con los
; nombres originales de esos parámetros. Obs: la macro 'xdefine',
; a diferencia de 'define', evita la redefinición.
%xdefine BUFFER_PTR    RDI
%xdefine END_OF_BUFFER RSI

;
; Vector h[8] donde se almacenará el hash calculado.
;
%xdefine H0 RDX
%xdefine H1 RDX + 4
%xdefine H2 RDX + 8
%xdefine H3 RDX + 12
%xdefine H4 RDX + 16
%xdefine H5 RDX + 20
%xdefine H6 RDX + 24
%xdefine H7 RDX + 28

;
; Utilizamos a registros de propósito general para mantener las
; variables A, B, C, D, E, F, G y H del algoritmo de hashing.
;
%xdefine A R8D
%xdefine B R9D
%xdefine C R10D
%xdefine D R11D
%xdefine E R12D
%xdefine F R13D
%xdefine G R14D
%xdefine H R15D

;
; Parámetros:
;
; %1:  número de ronda (de 0 a 63).
; %2:  puntero al buffer.
; %3:  A
; %4:  B
; %5:  C
; %6:  D
; %7:  E
; %8:  F
; %9:  G
; %10: H
;
%macro UPDATE_ABCDEFGH 10

  %if (%1 < 16)
    MOV EAX, [%2 + 4*%1]

    ;
    ; Formateamos el valor a 'Little Endian'.
    BSWAP EAX

  %elif
    ;
    ; Obtenemos 'sigma 0' (s0):
    ;
    ;   s0 = ROTATE_RIGHT(w[i-15], 7) ^ ROTATE_RIGHT(w[i-15], 18) ^ SHIFT_RIGHT(w[i-15], 3)
    ;
    ; La expresión puede reescribirse de la siguiente forma:
    ;
    ;   s0 = ROTATE_RIGHT(w[i-15] ^ ROTATE_RIGHT(w[i-15], 11), 7) ^ SHIFT_RIGHT(w[i-15], 3)
    ;
    ; Esto nos permite ahorrar un par de instrucciones.
    ;
    MOV EAX, [w + (%1-15)*4]  ; EAX = w[i-15]
    MOV EBX, EAX
    ROR EAX, 11
    XOR EAX, EBX
    ROR EAX, 7
    SHR EBX, 3
    XOR EAX, EBX  ; EAX = s0

    ;
    ; Obtenemos 'sigma 1' (s1):
    ;
    ;   s1 = ROTATE_RIGHT(w[i-2], 17) ^ ROTATE_RIGHT(w[i-2], 19) ^ SHIFT_RIGHT(w[i-2], 10)
    ;
    ; La expresión puede reescribirse de la siguiente forma:
    ;
    ;   s1 = ROTATE_RIGHT(w[i-2] ^ ROTATE_RIGHT(w[i-2], 2), 17) ^ SHIFT_RIGHT(w[i-2], 10)
    ;
    ; Esto nos permite ahorrar un par de instrucciones.
    ;
    MOV EBX, [w + (%1-2)*4]  ; EBX = w[i-2]
    MOV ECX, EBX
    ROR ECX, 2
    XOR ECX, EBX
    ROR ECX, 17
    SHR EBX, 10
    XOR ECX, EBX  ; ECX = s1

    ;
    ; Obtenemos finalmente el w[i]
    ;
    ; w[i] =  w[i-16] + s0 + w[i-7] + s1
    ;
    ADD EAX, ECX             ; EAX = s0 + s1
    ADD EAX, [w + (%1-16)*4] ; EAX = s0 + s1 + w[i-16]
    ADD EAX, [w + (%1-7)*4]  ; EAX = w[i] = s0 + s1 + w[i-16] + w[i-7]
  %endif

  ;
  ; Guardamos el w[i] obtenido en el vector.
  MOV [w + 4*%1], EAX  ; w[i] = EAX

  ;
  ; Obtenemos 'S1':
  ;
  ;   S1 = ROTATE_RIGHT(E, 6) ^ ROTATE_RIGHT(E, 11) ^ ROTATE_RIGHT(E, 25)
  ;
  ; Que se puede escribir:
  ;
  ;   S1 = ROTATE_RIGHT(E ^ ROTATE_RIGHT(E XOR ROTATE_RIGHT(E, 14), 5), 6)
  ;
  MOV EBX, %7  ; EBX = E
  ROR EBX, 14
  XOR EBX, %7
  ROR EBX, 5
  XOR EBX, %7
  ROR EBX, 6

  ADD EAX, EBX  ; EAX = w[i] + S1

  ;
  ; Obtenemos 'ch':
  ;
  ;  ch = (E and F) XOR (NOT(E) AND G)
  ;
  MOV EBX, %8   ; EBX = F
  AND EBX, %7   ; EBX = F AND E
  MOV ECX, %7   ; ECX = E
  NOT ECX       ; ECX = NOT(E)
  AND ECX, %9   ; ECX = NOT(E) AND G
  XOR EBX, ECX  ; EBX = (F AND E) XOR (NOT(E) AND G)

  ;
  ; Completamos el cálculo de temp1:
  ;
  ;  temp1 = w[i] + S1 + ch + H + k[i]
  ;
  ADD EAX, EBX         ; EAX = (w[i] + S1) + ch
  ADD EAX, %10         ; EAX = w[i] + S1 + ch + H
  ADD EAX, [k + %1*4]  ; EAX = w[i] + S1 + ch + H + k[i] = temp1

  ;
  ; Actualizamos las variables (D y H) que son afectadas por
  ; 'temp1', y de esta forma liberamos a EAX.
  ADD %6, EAX   ; D = D + temp1
  MOV %10, EAX  ; H = temp1

  ;
  ; Obtenemos 'S0':
  ;
  ;   S0 = ROTATE_RIGHT(A, 2) ^ ROTATE_RIGHT(A, 13) ^ ROTATE_RIGHT(A, 22)
  ;
  ; Que se puede escribir:
  ;
  ;   S0 = ROTATE_RIGHT(A ^ ROTATE_RIGHT(A XOR ROTATE_RIGHT(A, 9), 11), 2)
  ;
  MOV EAX, %3   ; EAX = A
  ROR EAX, 9
  XOR EAX, %3
  ROR EAX, 11
  XOR EAX, %3
  ROR EAX, 2    ; EAX = S0

  ADD %10, EAX  ; H = temp1 + S0

  ;
  ; Obtenemos 'maj':
  ;
  ;  maj = (A AND B) XOR (A AND C) XOR (B AND C)
  ;
  MOV EAX, %3   ; EAX = A
  MOV EBX, EAX  ; EBX = A
  AND EAX, %4   ; EAX = A AND B
  AND EBX, %5   ; EBX = A AND C
  XOR EAX, EBX  ; EAX = (A AND B) XOR (A AND C)
  MOV EBX, %4   ; EBX = B
  AND EBX, %5   ; EBX = B AND C
  XOR EAX, EBX  ; EAX = (A AND B) XOR (A AND C) XOR (B AND C)

  ADD %10, EAX  ; H = temp1 + S0 +  maj  (Obs: S0 + maj = temp2)

%endmacro

;------------------------------------------------------------------------
section .data
  DEFAULT REL

  ;
  ; Vector con las 64 constantes K: una para cada ronda. Estas
  ; constantes se pueden obtener de la siguiente forma: consiste
  ; cada una de los primeros 32 bits decimales de la raíz
  ; cuadrada de los primeros 64 números primos.
  ;
  k DD 0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5, 0x3956C25B, 0x59F111F1,\
       0x923F82A4, 0xAB1C5ED5, 0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3,\
       0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174, 0xE49B69C1, 0xEFBE4786,\
       0x0FC19DC6, 0x240CA1CC, 0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA,\
       0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7, 0xC6E00BF3, 0xD5A79147,\
       0x06CA6351, 0x14292967, 0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13,\
       0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85, 0xA2BFE8A1, 0xA81A664B,\
       0xC24B8B70, 0xC76C51A3, 0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070,\
       0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5, 0x391C0CB3, 0x4ED8AA4A,\
       0x5B9CCA4F, 0x682E6FF3, 0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208,\
       0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2
  ;
  ; Vector w de 64 DWORDs para los 'w'. (Obs: TIMES es el
  ; equivalente a DUP de MASM)
  ;
  w TIMES 64 DD 0

;------------------------------------------------------------------------
section .text

get_sha256_asm:

  PUSH RBP
  MOV RBP, RSP

  ; Guardamos los registros de la convención C.
  PUSH RBX
  PUSH R12
  PUSH R13
  PUSH R14
  PUSH R15

  ;
  ; Inicializaciones.
  ;
  ; H[0] = 0x6A09E667
  ; H[1] = 0xBB67AE85
  ; H[2] = 0x3C6EF372
  ; H[3] = 0xA54FF53A
  ; H[4] = 0x510E527F
  ; H[5] = 0x9B05688C
  ; H[6] = 0x1F83D9AB
  ; H[7] = 0x5BE0CD19
  ;
  MOV DWORD [H0], 0x6A09E667
  MOV DWORD [H1], 0xBB67AE85
  MOV DWORD [H2], 0x3C6EF372
  MOV DWORD [H3], 0xA54FF53A
  MOV DWORD [H4], 0x510E527F
  MOV DWORD [H5], 0x9B05688C
  MOV DWORD [H6], 0x1F83D9AB
  MOV DWORD [H7], 0x5BE0CD19

  ;
  ; A = H[0]
  ; B = H[1]
  ; C = H[2]
  ; D = H[3]
  ; E = H[4]
  ; F = H[5]
  ; G = H[6]
  ; H = H[7]
  ;
  MOV A, [H0]
  MOV B, [H1]
  MOV C, [H2]
  MOV D, [H3]
  MOV E, [H4]
  MOV F, [H5]
  MOV G, [H6]
  MOV H, [H7]

  ADD END_OF_BUFFER, BUFFER_PTR

loop:

  UPDATE_ABCDEFGH 0,  BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 1,  BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 2,  BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 3,  BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 4,  BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 5,  BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 6,  BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 7,  BUFFER_PTR, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 8,  BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 9,  BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 10, BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 11, BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 12, BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 13, BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 14, BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 15, BUFFER_PTR, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 16, BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 17, BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 18, BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 19, BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 20, BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 21, BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 22, BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 23, BUFFER_PTR, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 24, BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 25, BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 26, BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 27, BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 28, BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 29, BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 30, BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 31, BUFFER_PTR, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 32, BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 33, BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 34, BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 35, BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 36, BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 37, BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 38, BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 39, BUFFER_PTR, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 40, BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 41, BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 42, BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 43, BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 44, BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 45, BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 46, BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 47, BUFFER_PTR, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 48, BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 49, BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 50, BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 51, BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 52, BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 53, BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 54, BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 55, BUFFER_PTR, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 56, BUFFER_PTR, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 57, BUFFER_PTR, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 58, BUFFER_PTR, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 59, BUFFER_PTR, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 60, BUFFER_PTR, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 61, BUFFER_PTR, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 62, BUFFER_PTR, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 63, BUFFER_PTR, B, C, D, E, F, G, H, A

  ;
  ; Actualizamos el hash:
  ;
  ; H[0] = H[0] + A;
  ; H[1] = H[1] + B;
  ; H[2] = H[2] + C;
  ; H[3] = H[3] + D;
  ; H[4] = H[4] + E;
  ; H[5] = H[4] + F;
  ; H[6] = H[4] + G;
  ; H[7] = H[4] + H;
  ;
  ADD A, [H0]
  ADD B, [H1]
  ADD C, [H2]
  ADD D, [H3]
  ADD E, [H4]
  ADD F, [H5]
  ADD G, [H6]
  ADD H, [H7]

  MOV [H0], A
  MOV [H1], B
  MOV [H2], C
  MOV [H3], D
  MOV [H4], E
  MOV [H5], F
  MOV [H6], G
  MOV [H7], H

  ADD BUFFER_PTR, 64
  CMP BUFFER_PTR, END_OF_BUFFER
  JNE loop

  ;
  ;------------------------------------------------------------


  ;
  ; Restauramos los registros de la convención C.
  POP R15
  POP R14
  POP R13
  POP R12
  POP RBX
  POP RBP

  RET

;------------------------------------------------------------------------

