
/*****************************************************************
 *
 * SHA-256 sobre SIMD
 *
 * Programa principal. Permite correr el algoritmo SHA-256
 * empleando dos tipos de implementaciones:
 *
 * - Ensamblador Escalar
 * - Ensamblador Vectorizado (SIMD)
 *
 * Modo de uso: ./sha256 -opción ruta_al_archivo
 *
 * donde la opción incluye:
 *
 *        -a: ejecución de la implementación escalar en ASM
 *        -s: ejecución de la implementación vectorizada en ASM con SIMD
 *
 * Esta misma información se obtiene ejecutando el programa sin
 * parámetros.
 *
 * Autor: Pino, Jorge Daniel - JDanielPino@gmail.com
 *
 ****************************************************************/

#include <stdio.h>
#include <string.h>	/* strcmp(), memcpy(), memset() */
#include <stdlib.h>	/* malloc() */
#include <time.h>	/* clock(), CLOCKS_PER_SEC */


//
// Alias para los tipos
//
typedef unsigned int  uint;
typedef unsigned char uchar;
typedef unsigned long ulong;


//
// Prototipos
//
void display_usage();
long get_file_size(char* path_to_file);
uint get_padded_message_size(ulong data_size);
void pad_message(uchar* buffer, ulong buffer_size, ulong data_size);

//
// Definiciones externas
//
extern void get_sha256_asm(uchar* buffer, size_t size, uint h[8]);
extern void get_sha256_simd(uchar* buffer, size_t size, uint h[8]);


//
// Constantes.
//
#define ASM_OPTION_STR   "-a"
#define SIMD_OPTION_STR  "-s"

#define ASM_OPTION_VAL   0
#define SIMD_OPTION_VAL  1

#define MAX_FILE_SIZE 1073741824*3 /* 3 GB */

#define ASM_OUTPUT  "sha256_asm_out.dat"
#define SIMD_OUTPUT "sha256_simd_out.dat"

//
// Función principal
//
int main(int argc, char **argv) {
  char*   path_to_file = NULL;
  uint    option;
  uchar*  buffer = NULL;
  long    file_size = 0;
  clock_t execution_start = 0;
  clock_t execution_end = 0;
  double  num_clocks = 0;
  char*   output_file = NULL;
  uint    hash[8];

  // Si se ejecuta el programa sin los parámetros
  // correspondientes, se muestra en pantalla el modo de uso, y
  // finaliza.
  if (argc < 3) {
    display_usage();
    return 0;
  }

  // Comprobamos que la opción especificada sea válida.
  if (strcmp(argv[1], ASM_OPTION_STR) == 0) {
    option = ASM_OPTION_VAL;
  } else if (strcmp(argv[1], SIMD_OPTION_STR) == 0) {
    option = SIMD_OPTION_VAL;
  } else {
    printf("\nLa opción no es válida.\n");
    display_usage();
    return 0;
  }

  // Verificamos la existencia del archivo.
  file_size = get_file_size(argv[2]);

  if (file_size < 0) {
    fprintf(stderr, "\nEl archivo no existe o no se puede abrir.\n");
    display_usage();
    exit(1);
  }

  if (file_size > (long) MAX_FILE_SIZE) {
    fprintf(stderr, "\nEl supera el tamaño máximo admitido (%lu Bytes).\n\n", (long) MAX_FILE_SIZE);
    exit(1);
  }

  // Abrimos el archivo.
  FILE* fp = fopen(argv[2], "r");
  if (fp == NULL) {
    fprintf(stderr, "\nNo se pudo abrir el archivo.\n\n");
    exit(1);
  }

  // Obtenemos el tamaño de la totalidad de los datos,
  // incluyendo el 'padding' o relleno definido por el algoritmo.
  uint padded_size = get_padded_message_size(file_size);

  // Reservamos la memoria para cargar la totalidad del archivo
  // más el 'padding' o relleno.
  buffer = malloc(padded_size);

  // Si 'malloc' falló...
  if (buffer == NULL) {
    fprintf(stderr, "\nFallo de 'malloc'.\n\n");
    fclose(fp);
    exit(1);
  }

  // Leemos todo el archivo de un tiro. El valor '1' en el tercer
  // parámetro especifica la cantidad de elementos de tamaño
  // 'file_size' que vamos a leer del archivo. En este caso,
  // leemos un único elemento cuyo tamaño es precisamente el
  // tamaño total del archivo.
  fread(buffer, file_size, 1, fp);
  fclose(fp);

  pad_message(buffer, padded_size, file_size);

  printf("\n - Cálculo de hash SHA-256 mediante implementación ");

  switch(option) {
    case ASM_OPTION_VAL:
      execution_start = clock();
      printf("escalar en ASM:");

      // Obtenemos el hash de sha1 mediante la implementación
      // escalar en ASM.
      get_sha256_asm(buffer, padded_size, hash);

      execution_end = clock();
      output_file = ASM_OUTPUT;
      break;

    default:
      execution_start = clock();
      printf("vectorizada en ASM con SIMD:");

      // Obtenemos el hash de sha1 mediante la implementación
      // escalar en ASM.
      get_sha256_simd(buffer, padded_size, hash);

      execution_end = clock();
      output_file = SIMD_OUTPUT;
      break;
  }

  printf("\n\n   %08x%08x%08x%08x%08x%08x%08x%08x", hash[0], hash[1], hash[2], hash[3], hash[4], hash[5], hash[6], hash[7]);

  num_clocks = ((double)(execution_end - execution_start))/CLOCKS_PER_SEC;
  printf("\n\n - Tiempo de CPU (en segs): %f", num_clocks);
  printf("\n - Tamaño del archivo: %uB", file_size);

  free(buffer);

  fp = fopen(output_file, "a");

  if (fp != NULL) {
    // Escribimos los datos en el archivo ya convertidos en
    // ASCII.
    fprintf(fp, "%u; %f\n", file_size, num_clocks);

    printf("\n - Salida agregada al archivo '%s'\n\n", output_file);
    fclose(fp);
  } else {
    fprintf(stderr, "\n - Error al escribir los datos en el  archivo '%s'.\n\n", output_file);
  }
}
// Fin función principal //


//
// Muestra en pantalla el modo de uso de la aplicación.
//
void display_usage() {
  printf("\nModo de uso: ./sha256 -opción ruta_al_archivo");
  printf("\n\ndonde la opción incluye:");
  printf("\n   -a: ejecución de la implementación escalar en ASM.");
  printf("\n   -s: ejecución de la implementación vectorizada en ASM con SIMD.\n\n");
}

//
// Devuelve el tamaño de un archivo, o -1 si el archivo no
// existe o no se puede abrir.
//
long get_file_size(char* path_to_file) {
  long ret_val = -1;

  FILE* fp = fopen(path_to_file, "r");

  if (fp != NULL) {
    // Nos posicionamos al final del  archivo.
    fseek(fp, 0, SEEK_END);
    ret_val = ftell(fp);
    fclose(fp);
  }

  return ret_val;
}

//
// Devuelve el tamaño total que deberían tener los datos
// incluyendo el 'padding' o relleno.
//
uint get_padded_message_size(ulong data_size) {
  ulong new_size = data_size + 9;
  long  remainder = new_size % 64;

  if (remainder != 0) {
    new_size += 64 - remainder;
  }

  return new_size;
}

//
// El proceso de 'padding' o "relleno" tiene por objeto que la
// totalidad de los datos, incluyendo al 'padding' mismo, sea
// múltiplo de 64 bytes, dado que el hash se obtiene recorriendo
// y operando secuencialmente sobre cada uno de esos bloques de
// 64 bytes (o 512 bits).
// Afortunadamente, el proceso se simplifica al trabajar a nivel
// de byte (recuérdese que el algoritmo SHA-1 está definido a
// nivel de bit).
//
// En concreto, el proceso es el siguiente:
//
// 1) Por definición, a los datos se le agregarán 9 bytes (que se
//    describen más abajo): nuevo_tamaño = tamaño_datos + 9
// 2) Con el nuevo tamaño de los datos, se determina la cantidad
//    de bytes que aún faltarían agregar para obtener el próximo
//    valor múltiplo de 64 bytes.
// 3) Se extienden los datos a ese nuevo valor múltiplo de 64.
// 4) De los 9 bytes comentados anteriormente, uno es el valor
//    0x80; los otros 8 corresponden al tamaño en bits de los
//    datos originales (sin el 'padding').
// 5) El valor 0x80 se ubica inmediatamente a continuación de los
//    datos; los 8 restantes al final del bloque extendido.
// 6) Entre medio del byte 0x80 y los 8 con el tamaño en bits, se
//    rellena con '0x0'.
//
// Es decir:
//
// (Datos) + ('0x80') + ('0x0' + '0x0' + ... + '0x0') + (tamaño_datos_en_bits)
//
// La totalidad de ese bloque ya "rellenado" debe ser múltiplo de
// 64 bytes.
//
// Obs:
//
// 1) Puede darse el caso en que la suma de los 9 bytes ya sea
//    múltiplo de 64 bytes y, por lo tanto, no sea necesario
//    agregar los 0x0 intermedios.
// 2) Los 8 bytes correspondientes al tamaño en bits de los datos
//    originales se expresan en 'big endian'. Es decir, el byte
//    más significativo se ubica en la posición menos
//    significativa.
//
void pad_message(uchar* buffer, ulong buffer_size, ulong data_size) {
  // Byte 0x80 que se ubica a continuación de los datos.
  buffer[data_size] = (uchar) 0x80;

  // Rellenamos con ceros desde la posición 'data_size + 1' hasta
  // el comienzo de los últimos 8 bytes finales.
  ulong n = (buffer_size - 8) - (data_size + 1);

  memset(&buffer[data_size + 1], 0x0, n);

  // Completamos el 'padding' colocando el tamaño del mensaje
  // original (esto es, sin el 'padding') expresado en bits, en
  // los últimos 8 bytes del buffer. Este valor debe expresarse
  // formato 'Big Endian'. Para almacenar este valor de forma correcta,
  // reapuntamos al ulong mediante un puntero a uchar, y
  // en un ciclo 'for' copiamos los datos en el orden correcto.
  ulong size_in_bits = data_size * 8;

  uchar* p = (uchar*) &size_in_bits;
  uint i = 0;

  for (i = 0; i < 8; i++) {
    buffer[buffer_size - i - 1] = *(((uchar*) &size_in_bits) + i);
  }
}

