
;*****************************************************************
;
; SHA-256 sobre SIMD - Versión Ensamblador Vectorizada (SIMD)
;
; Pino, Jorge Daniel - JDanielPino@gmail.com
;
;*****************************************************************

; void get_sha256_simd(unsigned char* buffer, unsigned int size, unsigned int h[8]);
global get_sha256_simd
;
; Los parámetros de la función 'get_sha256_simd' se reciben en los
; siguientes registros:
;
; RDI:  'buffer'
; RSI:  'size'
; RDX:  'h[8]'
;

;----------------------------------------------------------------
;
; Definiciones de macros
;


;
; Asignamos aliases a los registros donde se reciben los
; parámetros, de manera tal que podamos referenciarlos con los
; nombres originales de esos parámetros. Obs: la macro 'xdefine',
; a diferencia de 'define', evita la redefinición.
%xdefine BUFFER_PTR    RDI
%xdefine END_OF_BUFFER RSI


;
; Vector h[8] donde se almacenará el hash calculado.
;
%xdefine H0 RDX
%xdefine H1 RDX + 4
%xdefine H2 RDX + 8
%xdefine H3 RDX + 12
%xdefine H4 RDX + 16
%xdefine H5 RDX + 20
%xdefine H6 RDX + 24
%xdefine H7 RDX + 28

;
; Utilizamos a registros de propósito general para mantener las
; variables A, B, C, D, E, F, G y H del algoritmo de hashing.
;
%xdefine A R8D
%xdefine B R9D
%xdefine C R10D
%xdefine D R11D
%xdefine E R12D
%xdefine F R13D
%xdefine G R14D
%xdefine H R15D


;
; Buffer circular compuesto por 4 registros XMM que almacenan
; 16 elementos (4 por cada registro XMM) del vector w definido en
; el algoritmo. Este buffer se utiliza para el precálculo de los
; w[i] y seguidamente la suma w[i] + k, que es, en definitiva, el
; valor que realmente importa en el algoritmo (utilizado para el
; cálculo de la variable 'a'). La propiedad 'circular' se refiere
; a que el grupo de elementos calculados más allá de W3 se
; almacenan comenzando nuevamente por W0. Por ejemplo, los w
; correspondientes a la ronda 16 (que son w[16], w[17], w[18] y
; w[19]) se almacenan en W0, los otros cuatro de la ronda 20, en
; w[20], y así sucesivamente.
;
; Importante!
; A efectos de almacenar y recuperar los valores en el orden
; correcto (más precisamente, en el cálculo vectorizado 'w + k'),
; cada registro XMM0 almacena una fila de elementos w en el orden
; inverso (w[j-3], w[j-2], w[j-1], w[j]). Es fundamental tener en
; cuenta esto a la hora de leer los comentarios a lo largo de
; todo el código.
;
%xdefine W0 XMM0  ; w[i+3]  | w[i+2] | w[i+1] | w[i]
%xdefine W1 XMM1  ; w[i+7]  |       ...       | w[i+4]
%xdefine W2 XMM2  ; w[i+11] |       ...       | w[i+8]
%xdefine W3 XMM3  ; w[i+15] |       ...       | w[i+12]

;
; Buffer circular análogo al anterior, que almacena la suma
; w[i] + k[i].
;
%xdefine WK0 XMM4  ; wk[i+3]  |     ...     | wk[i]
%xdefine WK1 XMM5  ; wk[i+7]  |     ...     | w[i+4]
%xdefine WK2 XMM6  ; wk[i+11] |     ...     | w[i+8]
%xdefine WK3 XMM7  ; wk[i+15] |     ...     | w[i+12]

;
; Definimos los registros XMM que podemos usar libremente.
%xdefine XMM_AUX0 XMM8
%xdefine XMM_AUX1 XMM9
%xdefine XMM_AUX2 XMM10
%xdefine XMM_AUX3 XMM11
%xdefine XMM_AUX4 XMM12
%xdefine XMM_AUX5 XMM13
%xdefine XMM_AUX6 XMM14

;
; Usamos este registro XMM para almacenar el formato de
; reordenamiento en 'Little Endian' de los bytes leídos desde
; memoria.
%xdefine LITTLE_ENDIAN_BYTES_SHUFFLING XMM15

;
; Esta macro centraliza las llamadas a las otras macros que
; calculan los elementos del vector WK de forma vectorizada.
%macro CALCULATE_WK 2

   %if (%1 < 16)
     CALCULATE_WK_FOR_ROUNDS_0_TO_15 %1, %2
   %else
     CALCULATE_WK_FOR_ROUNDS_16_TO_63 %1
   %endif

%endmacro

;
; Esta macro obtiene los primeros 16 'w' del algoritmo (de w[0] a
; w[15]), los almacena en el vector W0...W3 de XMMs, calcula los
; 'w[i] + k1' y los almacena en el vector 'wk' en memoria. Estos
; primeros 16 elementos se leen directamente del buffer y no
; requieren de mayor procesamiento. Parámetros:
;
; %1: número de ronda (solo válida para i = 0..15).
; %2: dirección del buffer.
;
; A los efectos de favorecer el mecanismo de 'pipelining' y la
; ejecución fuera de orden, la obtención de los cuatro elementos
; 'w' simultáneos usando SIMD se particiona en cuatro etapas,
; cada una de las cuales se distingue mediante el índice del
; número de ronda (primer parámetro). Esta partición en etapas
; favorece la intercalación con otras instrucciones
; independientes, y es precisamente esto lo que posibilita la
; ejecución fuera de orden.
;
%macro CALCULATE_WK_FOR_ROUNDS_0_TO_15 2

  ;
  ; Aplicamos una 'and' al índice 'i'. Esto nos permitirá
  ; distinguir el número de etapa.
  %assign i (%1 & 3)

  ;
  ; Etapa 1: rondas 0, 4, 8, 12. Simplemente leemos del buffer.
  ;
  %if (i == 0)

    ;
    ; Si es la primera ronda, empezar por el principio de vector
    ; armado con los registros XMM (W0, W1, W2, W3).
    %if (%1 == 0)
      %xdefine W  W0
      %xdefine WK WK0
    %endif

    ;
    ; Leeremos de a bloques de 16 bytes en cada acceso. Los
    ; offsets que deberemos generar cada vez serán los
    ; siguientes: BUFFER_PTR (ronda 0), BUFFER_PTR + 16
    ; (ronda 4), BUFFER_PTR + 32 (ronda 8) y BUFFER_PTR + 48
    ; (ronda 12). Eso es lo que hace la expresión debajo.
    MOVDQU W, [%2 + 4*%1]

  ;
  ; Etapa 2: rondas 1, 5, 9, 13. Ejecutamos el formateo de los
  ; datos.
  ;
  %elif (i == 1)
    PSHUFB W, LITTLE_ENDIAN_BYTES_SHUFFLING

  ;
  ; Etapa 3: rondas 2, 6, 10, 14. No tenemos instrucciones
  ; adicionales que podamos realizar. Por lo tanto, no hacemos
  ; nada.
  ;
  %elif (i == 2)
    ;
    ; No hacemos nada.
    ;
  ;
  ; Etapa 4: rondas 3, 7, 11, 15. A esta altura, la macro
  ; UPDATE_ABCDEFGH, encargada de de la actualización de las
  ; variables A, B, ..., H, ya consumió los WK[j] que ahora serán
  ; sobreescritos. Por lo tanto, incluir el código que sigue en
  ; una etapa anterio generaría un resultado incorrecto.
  ;
  %else
    MOVDQU WK, [k + (%1 - 3)*4]  ; WK = k[i+3] | k[i+2] | k[i+1] | k[i]
    PADDD  WK, W                 ; WK = k[i+3] + w[i+3] | ... | k[i] + w[i]

    ;
    ; Reasignamos el símbolo W para que apunte al siguiente
    ; registro XMM en el vector de w's que se está armando.
    %if (%1 == 3)
      %xdefine W  W1
      %xdefine WK WK1

    %elif (%1 == 7)
      %xdefine W  W2
      %xdefine WK WK2

    %elif (%1 == 11)
      %xdefine W  W3
      %xdefine WK WK3

    %endif
  %endif

%endmacro

;
; Esta macro obtiene los elementos 'w[16]' a 'w[63] del
; algoritmo, los almacena de manera circular en el buffer
; W0...W4 de XMMs, calcula los 'w[i] + k' y los
; almacena en otro buffer denominado 'WK'. Lo hace de forma
; vectorizada calculando de a 4 elementos a la vez. Con esta
; macro se inicia el mecanismo de 'pipelining' combinado con el
; método de optimización 'loop unrolling', en el cual un bucle se
; depliega para ahorrar las intrucciones de control del mismo.
;
; Parámetro:
;
; %1: número de ronda.
;
; A los efectos de favorecer el mecanismo de 'pipelining' y la
; ejecución fuera de orden, la obtención de los cuatro elementos
; 'w' simultáneos usando SIMD se particiona en cuatro etapas,
; cada una de las cuales se distingue mediante el índice del
; número de ronda (primer parámetro). Esta partición en etapas
; favorece la intercalación con otras instrucciones
; independientes, y es precisamente esto lo que posibilita la
; ejecución fuera de orden.
;
%macro CALCULATE_WK_FOR_ROUNDS_16_TO_63 1

  ;
  ; Aplicamos una 'and' al índice 'i'. Esto nos permitirá
  ; distinguir el número de etapa.
  %assign i (%1 & 3)

  ;
  ; Selector. Nos permitirá movernos de forma circular por los
  ; búfferes W  (w0, w1, w2 y w3) y WK (WK0, WK1, WK2 y WK3).
  %assign s (%1 & 12)/4

  ;
  ; Etapa 1: rondas 16, 20, 24, 28, 32, ... 60
  ;
  %if (i == 0)

    ;
    ; Seleccionamos los registros XMM que correspondan del buffer
    ; W mediante una reasignación de los símbolos.
    ;
    %if (s == 0)
      %xdefine W_MINUS_16  W0
      %xdefine W_MINUS_12  W1

    %elif (s == 1)
      %xdefine W_MINUS_16  W1
      %xdefine W_MINUS_12  W2

    %elif (s == 2)
      %xdefine W_MINUS_16  W2
      %xdefine W_MINUS_12  W3

    %else
      %xdefine W_MINUS_16  W3
      %xdefine W_MINUS_12  W0
    %endif

    ;
    ; Calculamos s0, definida como:
    ;
    ; s0 = (w[i-15] >>> 7) XOR (w[i-15] >>> 18) XOR (w[i-15] >> 2)
    ;
    ; Donde:
    ;
    ; >>>: right rotate
    ; >>:  right shift
    ;
    ; Considerando que en SSE3 no existe una instrucción para
    ; realizar rotaciones empaquetadas (esto es, de a varios
    ; elementos a la vez en los registros SIMD), reescribimos la
    ; expresión tal como figura a continuación. Nótese que este
    ; formato minimiza la cantidad de copias de w[i-15] en los
    ; registros:
    ;
    ; s0 = [(w[i-15] XOR (w[i-15] << 11) << 14] XOR
    ;      XOR  [(((w[i-15] XOR (w[i-15] >> 11)) >> 4) XOR w[i-15]) >> 3]
    ;
    MOVDQU  XMM_AUX1, W_MINUS_12      ; XMM_AUX1 = w[i-9]  | w[i-10] | w[i-11] | w[i-12]
    PALIGNR XMM_AUX1, W_MINUS_16, 4   ; XMM_AUX1 = w[i-12] | w[i-13] | w[i-14] | w[i-15]

    MOVDQU XMM_AUX0, XMM_AUX1  ; XMM_AUX0 = w[i-12] | w[i-13] | w[i-14] | w[i-15]
    PSLLD  XMM_AUX0, 11        ; XMM_AUX0 =   ...   |   w[i-15] << 11
    PXOR   XMM_AUX0, XMM_AUX1  ; XMM_AUX0 =   ...   |  (w[i-15] << 11) XOR w[i-15]
    PSLLD  XMM_AUX0, 14        ; XMM_AUX0 =   ...   | ((w[i-15] << 11) XOR w[i-15]) << 14

    MOVDQU XMM_AUX2, XMM_AUX1  ; XMM_AUX2 = w[i-12] | w[i-13] | w[i-14] | w[i-15]
    PSRLD  XMM_AUX2, 11        ; XMM_AUX2 =   ...   |   w[i-15] >> 11
    PXOR   XMM_AUX2, XMM_AUX1  ; XMM_AUX2 =   ...   |  (w[i-15] >> 11) XOR w[i-15]
    PSRLD  XMM_AUX2, 4         ; XMM_AUX2 =   ...   | ((w[i-15] << 11) XOR w[i-15]) >> 4 = A
    PXOR   XMM_AUX2, XMM_AUX1  ; XMM_AUX2 =   ...   |  A XOR w[i-15]
    PSRLD  XMM_AUX2, 3         ; XMM_AUX2 =   ...   | (A XOR w[i-15]) >> 3

    PXOR   XMM_AUX0, XMM_AUX2  ; XMM_AUX0 = s0[i-12] | s0[i-13] | s0[i-14] | s0[i-15]

  ;
  ; Etapa 2: rondas 17, 21, 25, 29, 33, ... 61
  ;
  %elif (i == 1)

    ;
    ; Seleccionamos los registros XMM que correspondan del buffer
    ; W mediante una reasignación de los símbolos.
    ;
    %if (s == 0)
      %xdefine W_MINUS_4  W3

    %elif (s == 1)
      %xdefine W_MINUS_4  W0

    %elif (s == 2)
      %xdefine W_MINUS_4  W1

    %else
      %xdefine W_MINUS_4  W2
    %endif

    ;
    ; Calculamos s1, definida como:
    ;
    ; s1 = (w[i-2] >>> 17) XOR (w[i-2] >>> 19) XOR (w[i-15] >> 10)
    ;
    ; Por los mismos motivos que s0, reescribimos la expresión:
    ;
    ; s1 = [(w[i-2] XOR (w[i-2] << 2)) << 13] XOR
    ;      XOR [(((w[i-2] XOR (w[i-2] >> 2)) >> 7) XOR w[i-2]) >> 10]
    ;
    ; - Obs:
    ; Nótese que el cómputo SIMD de 's1' debe limitarse a dos
    ; elementos: w[i-2] y w[i-1]. Los otros dos elementos
    ; siguientes (w[i] y w[i+1]) están en proceso de obtención.
    ; Esta situación desafortunada afecta considerablemente la
    ; performance de la aplicación, puesto que el cómputo de 's1'
    ; para los dos 's1' faltantes deberá repetirse.
    ;

    MOVDQU XMM_AUX1, W_MINUS_4  ; XMM_AUX1 = w[i-1] | w[i-2] | w[i-3] | w[i-4]
    PSLLD  XMM_AUX1, 2          ; XMM_AUX1 =   ...  |   w[i-2] << 2
    PXOR   XMM_AUX1, W_MINUS_4  ; XMM_AUX1 =   ...  |  (w[i-2] << 2) XOR w[i-2]
    PSLLD  XMM_AUX1, 13         ; XMM_AUX1 =   ...  | ((w[i-2] << 2) XOR w[i-2]) << 13

    MOVDQU XMM_AUX2, W_MINUS_4  ; XMM_AUX2 = w[i-1] | w[i-2] | ??? | ???
    PSRLD  XMM_AUX2, 2          ; XMM_AUX2 =   ...  |   w[i-2] >> 2 | ...
    PXOR   XMM_AUX2, W_MINUS_4  ; XMM_AUX2 =   ...  |  (w[i-2] >> 2) XOR w[i-2] | ...
    PSRLD  XMM_AUX2, 7          ; XMM_AUX2 =   ...  | ((w[i-2] << 2) XOR w[i-2]) >> 7 = A | ...
    PXOR   XMM_AUX2, W_MINUS_4  ; XMM_AUX2 =   ...  |  A XOR w[i-2] | ...
    PSRLD  XMM_AUX2, 10         ; XMM_AUX2 =   ...  | (A XOR w[i-2]) >> 10 | ...

    ;
    ; Obtenemos finalmente s0.
    PXOR XMM_AUX1, XMM_AUX2     ; XMM_AUX1 =   s1[i-1] | s1[i-2] |   ???   |   ???
    PSRLDQ XMM_AUX1, 8          ; XMM_AUX1 =     0x0   |   0x0   | s1[i-1] | s1[i-2]

  ;
  ; Etapa 3: rondas 18, 22, 26, 30, 34, ... 62
  ;
  %elif (i == 2)

    ;
    ; Seleccionamos los registros XMM que correspondan del buffer
    ; W mediante una reasignación de los símbolos.
    ;
    %if (s == 0)
      %xdefine W_MINUS_16  W0
      %xdefine W_MINUS_8   W2
      %xdefine W_MINUS_4   W3

    %elif (s == 1)
      %xdefine W_MINUS_16  W1
      %xdefine W_MINUS_8   W3
      %xdefine W_MINUS_4   W0

    %elif (s == 2)
      %xdefine W_MINUS_16  W2
      %xdefine W_MINUS_8   W0
      %xdefine W_MINUS_4   W1

    %else
      %xdefine W_MINUS_16  W3
      %xdefine W_MINUS_8   W1
      %xdefine W_MINUS_4   W2
    %endif

    MOVDQU  XMM_AUX2, W_MINUS_4
    PALIGNR XMM_AUX2, W_MINUS_8, 4  ; XMM_AUX2 = w[i-4] | w[i-5] | w[i-6] | w[i-7]

    ;
    ; Los w[i], w[i+1], w'[i+2] y w'[i+3] los obtendremos en el
    ; registro XMM_AUX0. Recuérdese:
    ;
    ; XMM_AUX0 = s0[i-12] | s0[i-13] | s0[i-14] | s0[i-15]
    ;
    PADDD XMM_AUX0, W_MINUS_16  ; XMM_AUX0 = ... | s0[i-15] + w[i-16]
    PADDD XMM_AUX0, XMM_AUX2    ; XMM_AUX0 = ... | s0[i-15] + w[i-16] + w[i-7]
    PADDD XMM_AUX0, XMM_AUX1    ; XMM_AUX0 = ... | s0[i-15] + w[i-16] + w[i-7] + s1[i-2]

    ;
    ; En concreto, tenemos lo siguiente:
    ;
    ; XMM_AUX0 = w'[i+3] | w'[i+2] | w[i+1] | w[i]
    ;

  ;
  ; Etapa 4: rondas 19, 23, 27, 31, 35, ... 63
  ;
  %else

    %if (s == 0)
      %xdefine W  W0
      %xdefine WK WK0

    %elif (s == 1)
      %xdefine W  W1
      %xdefine WK WK1

    %elif (s == 2)
      %xdefine W  W2
      %xdefine WK WK2

    %else
      %xdefine W  W3
      %xdefine WK WK3
    %endif

    MOVDQU W, XMM_AUX0   ;        W =  w'[i+3] | w'[i+2] | w[i+1] | w[i]
    MOVDQU XMM_AUX1, W   ; XMM_AUX1 =  w'[i+3] | w'[i+2] | w[i+1] | w[i]

    PSLLD  XMM_AUX1, 2   ; XMM_AUX1 =  ...  |   w[i] << 2
    PXOR   XMM_AUX1, W   ; XMM_AUX1 =  ...  |  (w[i] << 2) XOR w[i]
    PSLLD  XMM_AUX1, 13  ; XMM_AUX1 =  ...  | ((w[i] << 2) XOR w[i]) << 13

    PSRLD  XMM_AUX0, 2   ; XMM_AUX2 =  ...  |   w[i] >> 2
    PXOR   XMM_AUX0, W   ; XMM_AUX2 =  ...  |  (w[i] >> 2) XOR w[i]
    PSRLD  XMM_AUX0, 7   ; XMM_AUX2 =  ...  | ((w[i] << 2) XOR w[i]) >> 7 = A
    PXOR   XMM_AUX0, W   ; XMM_AUX2 =  ...  |  A XOR w[i]
    PSRLD  XMM_AUX0, 10  ; XMM_AUX2 =  ...  | (A XOR w[i]) >> 10

    ;
    ;
    ; Obtenemos finalmente s1 restantes.
    ;
    PXOR XMM_AUX1, XMM_AUX0  ; XMM_AUX1 =   |   ???    |    ???  | s1[i+1] | s1[i]
    PSLLDQ XMM_AUX1, 8       ; XMM_AUX1 =   |  s1[i+1] |  s1[i]  |   0x0   |  0x0

    PADDD  W, XMM_AUX1       ; W = w[i+3] | w[i+2] | w[i+1] w[i]

    MOVDQU WK, [k + (%1 - 19)*4 + 64]  ; Al sumar 64, saltamos los primeros 16 DWORDs
                                       ; correspondientes al rango k[0] a k[16].
    PADDD  WK, W

  %endif

%endmacro


;
; Parámetros:
;
; %1: número de ronda (de 0 a 63).
; %2: A
; %3: B
; %4: C
; %5: D
; %6: E
; %7: F
; %8: G
; %9: H
;
%macro UPDATE_ABCDEFGH 9

  ;
  ; Aplicamos una 'and' al índice 'i' para distinguir el selector
  ; (s) y el índice (i) del vector W. Ejemplo: si el número de
  ; ronda %1 vale 26, entonces el elemento se encontrará en W6[2]
  ; (esto es, el tercer elemento del W6 del buffer circular).
  ; Esta distinción se hace de la siguientes manera: consideramos
  ; únicamente los 5 bit menos significativos del número de ronda
  ; indicado por el parámetro %1 de la macro. De esos 5 bits, los
  ; 3 bits más significativos funcionan como selector de los
  ; vectores W (W0, W1, ... W7), mientras que los 2 bits menos
  ; significativos sirven de índice para ubicar finalmente al
  ; valor de los 4 DWORDs almacenados en el W (que es un registro
  ; XMM).
  %assign s (%1 & 12)/4  ; Selector del elemento WK del buffer (WK0, WK1, WK2, WK3).
  %assign i (%1 & 3)     ; Selector del elemento DWORD en el WK anterior.

  %if   (s == 0)
    %xdefine WK WK0

  %elif (s == 1)
    %xdefine WK WK1

  %elif (s == 2)
    %xdefine WK WK2

  %else
    %xdefine WK WK3

  %endif

  MOVDQU XMM_AUX6, WK

  ;
  ; Obtenemos finalmente el elemento w de interés.
  ;
  PSRLDQ XMM_AUX6, i*4  ; Desplazamos a derecha el WK tantos bytes como i*4
                        ; para poder guardarlo luego en EAX.

  MOVD EAX, XMM_AUX6    ; EAX = wk = w[i] + k[i]

  ;
  ; Obtenemos 'S1':
  ;
  ;   S1 = ROTATE_RIGHT(E, 6) ^ ROTATE_RIGHT(E, 11) ^ ROTATE_RIGHT(E, 25)
  ;
  ; Que puede escribirse:
  ;
  ;   S1 = ROTATE_RIGHT(E ^ ROTATE_RIGHT(E XOR ROTATE_RIGHT(E, 14), 5), 6)
  ;
  MOV EBX, %6  ; EBX = E
  ROR EBX, 14
  XOR EBX, %6
  ROR EBX, 5
  XOR EBX, %6
  ROR EBX, 6   ; EBX = S1

  ADD EAX, EBX  ; EAX = w[i] + k[i] + S1

  ;
  ; Obtenemos 'ch':
  ;
  ;  ch = (E and F) XOR (NOT(E) AND G)
  ;
  MOV EBX, %7   ; EBX = F
  AND EBX, %6   ; EBX = F AND E
  MOV ECX, %6   ; ECX = E
  NOT ECX       ; ECX = NOT(E)
  AND ECX, %8   ; ECX = NOT(E) AND G
  XOR EBX, ECX  ; EBX = (F AND E) XOR (NOT(E) AND G) = ch

  ;
  ; Completamos el cálculo de temp1:
  ;
  ;  temp1 = w[i] + S1 + ch + H + k[i]
  ;
  ADD EAX, EBX         ; EAX = (w[i] + k[i] + S1) + ch
  ADD EAX, %9          ; EAX = (w[i] + k[i] + S1) + ch + H = temp1

  ;
  ; Actualizamos las variables (D y H) que son afectadas por
  ; 'temp1', y de esta forma liberamos a EAX.
  ADD %5, EAX  ; D = D + temp1
  MOV %9, EAX  ; H = temp1

  ;
  ; Obtenemos 'S0':
  ;
  ;   S0 = ROTATE_RIGHT(A, 2) ^ ROTATE_RIGHT(A, 13) ^ ROTATE_RIGHT(A, 22)
  ;
  ; Que se puede escribir:
  ;
  ;   S0 = ROTATE_RIGHT(A ^ ROTATE_RIGHT(A XOR ROTATE_RIGHT(A, 9), 11), 2)
  ;
  MOV EAX, %2   ; EAX = A
  ROR EAX, 9
  XOR EAX, %2
  ROR EAX, 11
  XOR EAX, %2
  ROR EAX, 2    ; EAX = S0

  ADD %9, EAX   ; H = temp1 + S0

  ;
  ; Obtenemos 'maj':
  ;
  ;  maj = (A AND B) XOR (A AND C) XOR (B AND C)
  ;
  MOV EAX, %2   ; EAX = A
  MOV EBX, EAX  ; EBX = A
  AND EAX, %3   ; EAX = A AND B
  AND EBX, %4   ; EBX = A AND C
  XOR EAX, EBX  ; EAX = (A AND B) XOR (A AND C)
  MOV EBX, %3   ; EBX = B
  AND EBX, %4   ; EBX = B AND C
  XOR EAX, EBX  ; EAX = (A AND B) XOR (A AND C) XOR (B AND C)

  ADD %9, EAX   ; H = temp1 + S0 +  maj  (Obs: S0 + maj = temp2)

%endmacro

;------------------------------------------------------------------------
section .data align=16
  DEFAULT REL

  ;
  ; Vector con las 64 constantes K: una para cada ronda. Estas
  ; constantes se pueden obtener de la siguiente forma: consiste
  ; cada una de los primeros 32 bits decimales de la raíz
  ; cuadrada de los primeros 64 números primos.
  ;
  k DD 0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5, 0x3956C25B, 0x59F111F1,\
       0x923F82A4, 0xAB1C5ED5, 0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3,\
       0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174, 0xE49B69C1, 0xEFBE4786,\
       0x0FC19DC6, 0x240CA1CC, 0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA,\
       0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7, 0xC6E00BF3, 0xD5A79147,\
       0x06CA6351, 0x14292967, 0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13,\
       0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85, 0xA2BFE8A1, 0xA81A664B,\
       0xC24B8B70, 0xC76C51A3, 0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070,\
       0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5, 0x391C0CB3, 0x4ED8AA4A,\
       0x5B9CCA4F, 0x682E6FF3, 0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208,\
       0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2

  ;
  ; Cada secuencia de 16 bytes que se lee desde el buffer será
  ; cargada en los registros XMM en el mismo orden en que
  ; encuentran en dicho buffer. Esos bloques de 16 bytes en
  ; realidad se interpretan como 4 palabras DWORD. Por lo tanto,
  ; habrá que reorganizar los datos en formato 'Little Endian'
  ; antes de operar con ellos según lo indica el algoritmo. Para
  ; ello, podemos utilizar la instrucción PSHUFB, de la extensión
  ; SSE3. Esta instrucción utiliza índices que determinan dónde
  ; debe reubicarse cada uno de los 16 bytes de un registro XMM.
  ; La variable que sigue define el orden de reubicación para
  ; formatear correctamente los 4 DWORDs.
  little_endian_bytes_shuffling_values DB 0x03, 0x02, 0x01, 0x00,\
                                          0x07, 0x06, 0x05, 0x04,\
                                          0x0b, 0x0a, 0x09, 0x08,\
                                          0x0f, 0x0e, 0x0d, 0x0c

;------------------------------------------------------------------------
section .text

get_sha256_simd:

  PUSH RBP
  MOV RBP, RSP

  ; Guardamos los registros de la convención C.
  PUSH RBX
  PUSH R12
  PUSH R13
  PUSH R14
  PUSH R15

  ;
  ; Obtenemos el desplazamiento final del último bloque que debe
  ; procesarse.
  ADD END_OF_BUFFER, BUFFER_PTR

  ;
  ; Inicializamos los valores para el formateo en 'Little Endian'
  ; de los datos leídos del buffer en un registro XMM.
  MOVDQU LITTLE_ENDIAN_BYTES_SHUFFLING, [little_endian_bytes_shuffling_values]

  ;
  ; Inicializaciones.
  ;
  ; H[0] = 0x6A09E667
  ; H[1] = 0xBB67AE85
  ; H[2] = 0x3C6EF372
  ; H[3] = 0xA54FF53A
  ; H[4] = 0x510E527F
  ; H[5] = 0x9B05688C
  ; H[6] = 0x1F83D9AB
  ; H[7] = 0x5BE0CD19
  ;
  MOV DWORD [H0], 0x6A09E667
  MOV DWORD [H1], 0xBB67AE85
  MOV DWORD [H2], 0x3C6EF372
  MOV DWORD [H3], 0xA54FF53A
  MOV DWORD [H4], 0x510E527F
  MOV DWORD [H5], 0x9B05688C
  MOV DWORD [H6], 0x1F83D9AB
  MOV DWORD [H7], 0x5BE0CD19

  ;
  ; Calculamos por adelantado 16 elementos del vector W (64
  ; bytes).
  CALCULATE_WK 0, BUFFER_PTR
  CALCULATE_WK 1, BUFFER_PTR
  CALCULATE_WK 2, BUFFER_PTR
  CALCULATE_WK 3, BUFFER_PTR  ; Obtenemos W0: w[0] | w[1] | w[2] | w[3]

  CALCULATE_WK 4, BUFFER_PTR
  CALCULATE_WK 5, BUFFER_PTR
  CALCULATE_WK 6, BUFFER_PTR
  CALCULATE_WK 7, BUFFER_PTR  ; Obtenemos W1: w[4] | w[5] | w[6] | w[7]

  CALCULATE_WK 8,  BUFFER_PTR
  CALCULATE_WK 9,  BUFFER_PTR
  CALCULATE_WK 10, BUFFER_PTR
  CALCULATE_WK 11, BUFFER_PTR  ; Obtenemos W2: w[8] | w[9] | w[10] | w[11]

  CALCULATE_WK 12, BUFFER_PTR
  CALCULATE_WK 13, BUFFER_PTR
  CALCULATE_WK 14, BUFFER_PTR
  CALCULATE_WK 15, BUFFER_PTR  ; Obtenemos W3: w[12] | w[13] | w[14] | w[15]


  ;
  ; A = H[0]
  ; B = H[1]
  ; C = H[2]
  ; D = H[3]
  ; E = H[4]
  ; F = H[5]
  ; G = H[6]
  ; H = H[7]
  ;
  MOV A, [H0]
  MOV B, [H1]
  MOV C, [H2]
  MOV D, [H3]
  MOV E, [H4]
  MOV F, [H5]
  MOV G, [H6]
  MOV H, [H7]

loop:

  UPDATE_ABCDEFGH 0,  A, B, C, D, E, F, G, H
  CALCULATE_WK 16, BUFFER_PTR

  UPDATE_ABCDEFGH 1,  H, A, B, C, D, E, F, G
  CALCULATE_WK 17, BUFFER_PTR

  UPDATE_ABCDEFGH 2,  G, H, A, B, C, D, E, F
  CALCULATE_WK 18, BUFFER_PTR

  UPDATE_ABCDEFGH 3,  F, G, H, A, B, C, D, E
  CALCULATE_WK 19, BUFFER_PTR     ; Obtenemos w[16], w[17], w[18] y w[19].

  UPDATE_ABCDEFGH 4,  E, F, G, H, A, B, C, D
  CALCULATE_WK 20, BUFFER_PTR

  UPDATE_ABCDEFGH 5,  D, E, F, G, H, A, B, C
  CALCULATE_WK 21, BUFFER_PTR

  UPDATE_ABCDEFGH 6,  C, D, E, F, G, H, A, B
  CALCULATE_WK 22, BUFFER_PTR

  UPDATE_ABCDEFGH 7,  B, C, D, E, F, G, H, A
  CALCULATE_WK 23, BUFFER_PTR     ; Obtenemos w[20], w[21], w[22] y w[23].

  UPDATE_ABCDEFGH 8,  A, B, C, D, E, F, G, H
  CALCULATE_WK 24, BUFFER_PTR

  UPDATE_ABCDEFGH 9,  H, A, B, C, D, E, F, G
  CALCULATE_WK 25, BUFFER_PTR

  UPDATE_ABCDEFGH 10, G, H, A, B, C, D, E, F
  CALCULATE_WK 26, BUFFER_PTR

  UPDATE_ABCDEFGH 11, F, G, H, A, B, C, D, E
  CALCULATE_WK 27, BUFFER_PTR     ; Obtenemos w[24], w[25], w[26] y w[27].

  UPDATE_ABCDEFGH 12, E, F, G, H, A, B, C, D
  CALCULATE_WK 28, BUFFER_PTR

  UPDATE_ABCDEFGH 13, D, E, F, G, H, A, B, C
  CALCULATE_WK 29, BUFFER_PTR

  UPDATE_ABCDEFGH 14, C, D, E, F, G, H, A, B
  CALCULATE_WK 30, BUFFER_PTR

  UPDATE_ABCDEFGH 15, B, C, D, E, F, G, H, A
  CALCULATE_WK 31, BUFFER_PTR     ; Obtenemos w[28], w[29], w[30] y w[31].

  UPDATE_ABCDEFGH 16, A, B, C, D, E, F, G, H
  CALCULATE_WK 32, BUFFER_PTR

  UPDATE_ABCDEFGH 17, H, A, B, C, D, E, F, G
  CALCULATE_WK 33, BUFFER_PTR

  UPDATE_ABCDEFGH 18, G, H, A, B, C, D, E, F
  CALCULATE_WK 34, BUFFER_PTR

  UPDATE_ABCDEFGH 19, F, G, H, A, B, C, D, E
  CALCULATE_WK 35, BUFFER_PTR     ; Obtenemos w[32], w[33], w[34] y w[35].

  UPDATE_ABCDEFGH 20, E, F, G, H, A, B, C, D
  CALCULATE_WK 36, BUFFER_PTR

  UPDATE_ABCDEFGH 21, D, E, F, G, H, A, B, C
  CALCULATE_WK 37, BUFFER_PTR

  UPDATE_ABCDEFGH 22, C, D, E, F, G, H, A, B
  CALCULATE_WK 38, BUFFER_PTR

  UPDATE_ABCDEFGH 23, B, C, D, E, F, G, H, A
  CALCULATE_WK 39, BUFFER_PTR     ; Obtenemos w[36], w[37], w[38] y w[39].

  UPDATE_ABCDEFGH 24, A, B, C, D, E, F, G, H
  CALCULATE_WK 40, BUFFER_PTR

  UPDATE_ABCDEFGH 25, H, A, B, C, D, E, F, G
  CALCULATE_WK 41, BUFFER_PTR

  UPDATE_ABCDEFGH 26, G, H, A, B, C, D, E, F
  CALCULATE_WK 42, BUFFER_PTR

  UPDATE_ABCDEFGH 27, F, G, H, A, B, C, D, E
  CALCULATE_WK 43, BUFFER_PTR     ; Obtenemos w[40], w[41], w[42] y w[43].

  UPDATE_ABCDEFGH 28, E, F, G, H, A, B, C, D
  CALCULATE_WK 44, BUFFER_PTR

  UPDATE_ABCDEFGH 29, D, E, F, G, H, A, B, C
  CALCULATE_WK 45, BUFFER_PTR

  UPDATE_ABCDEFGH 30, C, D, E, F, G, H, A, B
  CALCULATE_WK 46, BUFFER_PTR

  UPDATE_ABCDEFGH 31, B, C, D, E, F, G, H, A
  CALCULATE_WK 47, BUFFER_PTR     ; Obtenemos w[44], w[45], w[46] y w[47].

  UPDATE_ABCDEFGH 32, A, B, C, D, E, F, G, H
  CALCULATE_WK 48, BUFFER_PTR

  UPDATE_ABCDEFGH 33, H, A, B, C, D, E, F, G
  CALCULATE_WK 49, BUFFER_PTR

  UPDATE_ABCDEFGH 34, G, H, A, B, C, D, E, F
  CALCULATE_WK 50, BUFFER_PTR

  UPDATE_ABCDEFGH 35, F, G, H, A, B, C, D, E
  CALCULATE_WK 51, BUFFER_PTR     ; Obtenemos w[48], w[49], w[50] y w[51].

  UPDATE_ABCDEFGH 36, E, F, G, H, A, B, C, D
  CALCULATE_WK 52, BUFFER_PTR

  UPDATE_ABCDEFGH 37, D, E, F, G, H, A, B, C
  CALCULATE_WK 53, BUFFER_PTR

  UPDATE_ABCDEFGH 38, C, D, E, F, G, H, A, B
  CALCULATE_WK 54, BUFFER_PTR

  UPDATE_ABCDEFGH 39, B, C, D, E, F, G, H, A
  CALCULATE_WK 55, BUFFER_PTR     ; Obtenemos w[52], w[53], w[54] y w[55].

  UPDATE_ABCDEFGH 40, A, B, C, D, E, F, G, H
  CALCULATE_WK 56, BUFFER_PTR

  UPDATE_ABCDEFGH 41, H, A, B, C, D, E, F, G
  CALCULATE_WK 57, BUFFER_PTR

  UPDATE_ABCDEFGH 42, G, H, A, B, C, D, E, F
  CALCULATE_WK 58, BUFFER_PTR

  UPDATE_ABCDEFGH 43, F, G, H, A, B, C, D, E
  CALCULATE_WK 59, BUFFER_PTR     ; Obtenemos w[56], w[57], w[58] y w[59].

  UPDATE_ABCDEFGH 44, E, F, G, H, A, B, C, D
  CALCULATE_WK 60, BUFFER_PTR

  UPDATE_ABCDEFGH 45, D, E, F, G, H, A, B, C
  CALCULATE_WK 61, BUFFER_PTR

  UPDATE_ABCDEFGH 46, C, D, E, F, G, H, A, B
  CALCULATE_WK 62, BUFFER_PTR

  UPDATE_ABCDEFGH 47, B, C, D, E, F, G, H, A
  CALCULATE_WK 63, BUFFER_PTR     ; Obtenemos w[60], w[61], w[52] y w[63].

  ;
  ; Nos deplazamos al siguiente bloque del buffer sumando 64
  ; bytes al puntero. Si este era el último, saltamos a la
  ; actualización de las variables A, B, C, D, E, F, G y H
  ; sin calcular por adelantado ningún otro elemento del
  ; vector W.
  ADD BUFFER_PTR, 64
  CMP BUFFER_PTR, END_OF_BUFFER
  JE last_abcdefgh_updates

  ;
  ; Comienza el cálculo por adelantado de los próximos 16
  ; elementos del vector W y los w + k, que se utilizarán para
  ; actualizar los valores de las variables A, B, C, D, E, F,
  ; G y H en la próxima iteración.
  ;
  UPDATE_ABCDEFGH 48, A, B, C, D, E, F, G, H
  CALCULATE_WK 0, BUFFER_PTR

  UPDATE_ABCDEFGH 49, H, A, B, C, D, E, F, G
  CALCULATE_WK 1, BUFFER_PTR

  UPDATE_ABCDEFGH 50, G, H, A, B, C, D, E, F
  CALCULATE_WK 2, BUFFER_PTR

  UPDATE_ABCDEFGH 51, F, G, H, A, B, C, D, E
  CALCULATE_WK 3, BUFFER_PTR     ; Obtenemos w[0], w[1], w[2] y w[3].

  UPDATE_ABCDEFGH 52, E, F, G, H, A, B, C, D
  CALCULATE_WK 4, BUFFER_PTR

  UPDATE_ABCDEFGH 53, D, E, F, G, H, A, B, C
  CALCULATE_WK 5, BUFFER_PTR

  UPDATE_ABCDEFGH 54, C, D, E, F, G, H, A, B
  CALCULATE_WK 6, BUFFER_PTR

  UPDATE_ABCDEFGH 55, B, C, D, E, F, G, H, A
  CALCULATE_WK 7, BUFFER_PTR     ; Obtenemos w[4], w[5], w[6] y w[7].

  UPDATE_ABCDEFGH 56, A, B, C, D, E, F, G, H
  CALCULATE_WK 8, BUFFER_PTR

  UPDATE_ABCDEFGH 57, H, A, B, C, D, E, F, G
  CALCULATE_WK 9, BUFFER_PTR

  UPDATE_ABCDEFGH 58, G, H, A, B, C, D, E, F
  CALCULATE_WK 10, BUFFER_PTR

  UPDATE_ABCDEFGH 59, F, G, H, A, B, C, D, E
  CALCULATE_WK 11, BUFFER_PTR     ; Obtenemos w[8], w[9], w[10] y w[11].

  UPDATE_ABCDEFGH 60, E, F, G, H, A, B, C, D
  CALCULATE_WK 12, BUFFER_PTR

  UPDATE_ABCDEFGH 61, D, E, F, G, H, A, B, C
  CALCULATE_WK 13, BUFFER_PTR

  UPDATE_ABCDEFGH 62, C, D, E, F, G, H, A, B
  CALCULATE_WK 14, BUFFER_PTR

  UPDATE_ABCDEFGH 63, B, C, D, E, F, G, H, A
  CALCULATE_WK 15, BUFFER_PTR     ; Obtenemos w[12], w[13], w[14] y w[15].

  ;
  ; Actualizamos el hash:
  ;
  ; H[0] = H[0] + A;
  ; H[1] = H[1] + B;
  ; H[2] = H[2] + C;
  ; H[3] = H[3] + D;
  ; H[4] = H[4] + E;
  ; H[5] = H[5] + F;
  ; H[6] = H[6] + G;
  ; H[7] = H[7] + H;
  ;
  ADD A, [H0]
  ADD B, [H1]
  ADD C, [H2]
  ADD D, [H3]
  ADD E, [H4]
  ADD F, [H5]
  ADD G, [H6]
  ADD H, [H7]

  MOV [H0], A
  MOV [H1], B
  MOV [H2], C
  MOV [H3], D
  MOV [H4], E
  MOV [H5], F
  MOV [H6], G
  MOV [H7], H

  ;
  ; Iteramos
  JMP loop
  ;
  ;------------------------------------------------------------

;
; Última actualización.
;
last_abcdefgh_updates:

  UPDATE_ABCDEFGH 48, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 49, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 50, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 51, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 52, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 53, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 54, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 55, B, C, D, E, F, G, H, A
  UPDATE_ABCDEFGH 56, A, B, C, D, E, F, G, H
  UPDATE_ABCDEFGH 57, H, A, B, C, D, E, F, G
  UPDATE_ABCDEFGH 58, G, H, A, B, C, D, E, F
  UPDATE_ABCDEFGH 59, F, G, H, A, B, C, D, E
  UPDATE_ABCDEFGH 60, E, F, G, H, A, B, C, D
  UPDATE_ABCDEFGH 61, D, E, F, G, H, A, B, C
  UPDATE_ABCDEFGH 62, C, D, E, F, G, H, A, B
  UPDATE_ABCDEFGH 63, B, C, D, E, F, G, H, A

  ;
  ; Actualizamos por última vez el hash:
  ;
  ; H[0] = H[0] + A;
  ; H[1] = H[1] + B;
  ; H[2] = H[2] + C;
  ; H[3] = H[3] + D;
  ; H[4] = H[4] + E;
  ; H[5] = H[5] + F;
  ; H[6] = H[6] + G;
  ; H[7] = H[7] + H;
  ;
  ADD A, [H0]
  ADD B, [H1]
  ADD C, [H2]
  ADD D, [H3]
  ADD E, [H4]
  ADD F, [H5]
  ADD G, [H6]
  ADD H, [H7]

  MOV [H0], A
  MOV [H1], B
  MOV [H2], C
  MOV [H3], D
  MOV [H4], E
  MOV [H5], F
  MOV [H6], G
  MOV [H7], H

  ;
  ;------------------------------------------------------------

  ;
  ; Restauramos los registros de la convención C.
  POP R15
  POP R14
  POP R13
  POP R12
  POP RBX
  POP RBP

  RET

;------------------------------------------------------------------------

